"""
Module to provide for an encapsulation of the table header markdown token.
"""

from typing import List, Optional, Union, cast

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.leaf_blocks.table_block_tuple import TableRow
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class TableMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the main table markdown token.
    """

    def __init__(
        self,
        position_marker: PositionMarker,
    ) -> None:
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_main,
            extra_data=None,
            position_marker=position_marker,
            extracted_whitespace="",
            requires_end_token=True,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_main

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownToken,
            TableMarkdownToken.__rehydrate_table_start,
            TableMarkdownToken.__rehydrate_table_end,
        )

    @staticmethod
    def __rehydrate_table_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return ""

    @staticmethod
    def __rehydrate_table_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, next_token, context)
        return ""

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownToken,
            TableMarkdownToken.__handle_start_table_token,
            TableMarkdownToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        token_parts: List[str] = []
        if (
            not output_html
            and transform_state.transform_stack
            and transform_state.transform_stack[-1].endswith("<li>")
        ):
            token_parts.append(ParserHelper.newline_character)
        token_parts.extend([output_html, "<table>", ParserHelper.newline_character])
        return "".join(token_parts)

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (next_token, transform_state)

        return "".join([output_html, "</table>", ParserHelper.newline_character])


class TableMarkdownHeaderToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the table header markdown token.
    """

    def __init__(
        self,
        header_table_row: TableRow,
        separator_table_row: TableRow,
        position_marker: PositionMarker,
    ) -> None:
        self.__header_row_leading_whitespace: Optional[str] = None
        self.__header_row_trailing_whitespace: Optional[str] = None
        self.__did_header_row_start_with_separator = False
        self.__separator_line: Optional[str] = None
        if header_table_row:
            self.__header_row_leading_whitespace = header_table_row.extracted_whitespace
            self.__header_row_trailing_whitespace = header_table_row.trailing_whitespace
            self.__did_header_row_start_with_separator = (
                header_table_row.did_start_with_separator
            )
            self.__separator_line = TableMarkdownHeaderToken.__xx(separator_table_row)

        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_header,
            extra_data=self.__compose_extra_data_field(),
            position_marker=position_marker,
            extracted_whitespace="",
            requires_end_token=True,
        )

    def __compose_extra_data_field(self) -> Optional[str]:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        if self.__header_row_leading_whitespace is None:
            self._set_extra_data(None)
            return None

        assert self.__header_row_trailing_whitespace is not None
        assert self.__header_row_trailing_whitespace is not None
        assert self.__separator_line is not None
        composed_field = MarkdownToken.extra_data_separator.join(
            [
                self.__header_row_leading_whitespace,
                self.__header_row_trailing_whitespace,
                str(self.__did_header_row_start_with_separator),
                self.__separator_line,
            ]
        )
        self._set_extra_data(composed_field)
        return composed_field

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_header

    # pylint: enable=protected-access

    @property
    def header_row_leading_whitespace(self) -> str:
        """Gets any leading whitespace for the header row."""
        assert self.__header_row_leading_whitespace is not None
        return self.__header_row_leading_whitespace

    @property
    def header_row_trailing_whitespace(self) -> str:
        """Gets any trailing whitespace for the header row."""
        assert self.__header_row_trailing_whitespace is not None
        return self.__header_row_trailing_whitespace

    @property
    def did_header_row_start_with_separator(self) -> bool:
        """Gets a value indicating whether the header row started with a separator."""
        return self.__did_header_row_start_with_separator

    @property
    def separator_line(self) -> str:
        """Gets the separator line between the header and the body of the table."""
        assert self.__separator_line is not None
        return self.__separator_line

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "header_row_leading_whitespace" and isinstance(
            field_value, str
        ):
            self.__header_row_leading_whitespace = field_value
            extra_data = self.__compose_extra_data_field()
            super()._set_extra_data(extra_data)
            return True
        if field_name == "separator_line" and isinstance(field_value, str):
            self.__separator_line = field_value
            extra_data = self.__compose_extra_data_field()
            super()._set_extra_data(extra_data)
            return True
        return super()._modify_token(field_name, field_value)

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownHeaderToken,
            TableMarkdownHeaderToken.__rehydrate_table_header_start,
            TableMarkdownHeaderToken.__rehydrate_table_header_end,
        )

    @staticmethod
    def __rehydrate_table_header_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = previous_token
        context.block_stack.append(current_token)

        header_start_text = cast(
            TableMarkdownHeaderToken, current_token
        ).header_row_leading_whitespace
        assert header_start_text is not None
        return header_start_text + (
            "|"
            if cast(
                TableMarkdownHeaderToken, current_token
            ).did_header_row_start_with_separator
            else ""
        )

    @staticmethod
    def __xx(separator_table_row: TableRow) -> str:
        header_end_text = separator_table_row.extracted_whitespace
        header_end_text += "|" if separator_table_row.did_start_with_separator else ""
        for i in separator_table_row.columns:
            header_end_text += i.leading_whitespace + i.text + i.trailing_whitespace
        header_end_text += separator_table_row.trailing_whitespace
        return header_end_text

    @staticmethod
    def __rehydrate_table_header_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, next_token, context)
        del context.block_stack[-1]
        start_token = cast(
            TableMarkdownHeaderToken,
            cast(EndMarkdownToken, current_token).start_markdown_token,
        )

        header_end_text = start_token.separator_line
        assert start_token.header_row_trailing_whitespace is not None
        assert header_end_text is not None
        return (
            start_token.header_row_trailing_whitespace + "\n" + header_end_text + "\n"
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownHeaderToken,
            TableMarkdownHeaderToken.__handle_start_table_token,
            TableMarkdownHeaderToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        return "".join(
            [
                output_html,
                "<thead>",
                ParserHelper.newline_character,
                "<tr>",
                ParserHelper.newline_character,
            ]
        )

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (next_token, transform_state)

        return "".join(
            [
                output_html,
                "</tr>",
                ParserHelper.newline_character,
                "</thead>",
                ParserHelper.newline_character,
            ]
        )


class TableMarkdownHeaderItemToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the table header item markdown token.
    """

    def __init__(
        self,
        leading_whitespace: str,
        column_alignment: Optional[str],
        line_number: int,
        column_number: int,
    ) -> None:
        self.__leading_whitespace: str = leading_whitespace
        self.__column_alignment = column_alignment
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_header_item,
            extra_data=self.__compose_extra_data_field(),
            line_number=line_number,
            column_number=column_number,
            extracted_whitespace=leading_whitespace,
            requires_end_token=True,
        )

    @property
    def column_alignment(self) -> Optional[str]:
        """Gets the alignment of this header's column."""
        return self.__column_alignment

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_header_item

    # pylint: enable=protected-access

    def __compose_extra_data_field(self) -> Optional[str]:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        composed_field = MarkdownToken.extra_data_separator.join(
            [
                self.__leading_whitespace,
                self.__column_alignment if self.__column_alignment is not None else "",
            ]
        )
        self._set_extra_data(composed_field)
        return composed_field

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownHeaderItemToken,
            TableMarkdownHeaderItemToken.__rehydrate_table_header_item_start,
            TableMarkdownHeaderItemToken.__rehydrate_table_header_item_end,
        )

    @staticmethod
    def __rehydrate_table_header_item_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return cast(TableMarkdownHeaderItemToken, current_token).extracted_whitespace

    @staticmethod
    def __rehydrate_table_header_item_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, next_token, context)
        return cast(EndMarkdownToken, current_token).extracted_whitespace

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownHeaderItemToken,
            TableMarkdownHeaderItemToken.__handle_start_table_token,
            TableMarkdownHeaderItemToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        table_token = cast(TableMarkdownHeaderItemToken, next_token)
        if table_token.column_alignment:
            text_to_add = f'<th align="{table_token.column_alignment}">'
        else:
            text_to_add = "<th>"
        return "".join(
            [
                output_html,
                text_to_add,
            ]
        )

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (next_token, transform_state)

        return "".join(
            [
                output_html,
                "</th>",
                ParserHelper.newline_character,
            ]
        )


class TableMarkdownBodyToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the table header markdown token.
    """

    def __init__(self, line_number: int, column_number: int) -> None:
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_body,
            extra_data=None,
            line_number=line_number,
            column_number=column_number,
            extracted_whitespace="",
            requires_end_token=True,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_body

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownBodyToken,
            TableMarkdownBodyToken.__rehydrate_table_body_token_start,
            TableMarkdownBodyToken.__rehydrate_table_body_token_end,
        )

    @staticmethod
    def __rehydrate_table_body_token_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return ""

    @staticmethod
    def __rehydrate_table_body_token_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, next_token, context)
        return ""

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownBodyToken,
            TableMarkdownBodyToken.__handle_start_table_token,
            TableMarkdownBodyToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        return "".join(
            [
                output_html,
                "<tbody>",
                ParserHelper.newline_character,
            ]
        )

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (next_token, transform_state)

        return "".join([output_html, "</tbody>", ParserHelper.newline_character])


class TableMarkdownRowToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the table header markdown token.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        leading_whitespace: str,
        trailing_whitespace: str,
        did_start_with_separator: bool,
        delta: int,
        line_number: int,
        column_number: int,
    ) -> None:
        self.__leading_whitespace = leading_whitespace
        self.__trailing_whitespace = trailing_whitespace
        self.__did_start_with_separator = did_start_with_separator
        self.__delta = delta
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_row,
            extra_data=self.__compose_extra_data_field(),
            line_number=line_number,
            column_number=column_number,
            extracted_whitespace="",
            requires_end_token=True,
        )

    # pylint: enable=too-many-arguments

    def __compose_extra_data_field(self) -> Optional[str]:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        composed_field = MarkdownToken.extra_data_separator.join(
            [
                self.__leading_whitespace,
                self.__trailing_whitespace,
                str(self.__did_start_with_separator),
                str(self.__delta),
            ]
        )
        self._set_extra_data(composed_field)
        return composed_field

    @property
    def leading_whitespace(self) -> str:
        """Gets any leading whitespace for this row."""
        return self.__leading_whitespace

    @property
    def trailing_whitespace(self) -> str:
        """Gets any trailing whitespace for this row."""
        return self.__trailing_whitespace

    @property
    def did_start_with_separator(self) -> bool:
        """Gets a value indicating whether the row started with a separator character."""
        return self.__did_start_with_separator

    @property
    def delta(self) -> int:
        """Gets the delta columns to add to ensure that this row has the required number of columns."""
        return self.__delta

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "leading_whitespace" and isinstance(field_value, str):
            self.__leading_whitespace = field_value
            extra_data = self.__compose_extra_data_field()
            super()._set_extra_data(extra_data)
            return True
        return super()._modify_token(field_name, field_value)

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_row

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownRowToken,
            TableMarkdownRowToken.__rehydrate_table_row_start,
            TableMarkdownRowToken.__rehydrate_table_row_end,
        )

    @staticmethod
    def __rehydrate_table_row_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        context.block_stack.append(current_token)
        return cast(TableMarkdownRowToken, current_token).leading_whitespace + (
            "|"
            if cast(TableMarkdownRowToken, current_token).did_start_with_separator
            else ""
        )

    @staticmethod
    def __rehydrate_table_row_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, next_token, context)
        del context.block_stack[-1]
        return (
            cast(EndMarkdownToken, current_token).extracted_whitespace
            + cast(
                TableMarkdownRowToken,
                cast(EndMarkdownToken, current_token).start_markdown_token,
            ).trailing_whitespace
            + "\n"
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownRowToken,
            TableMarkdownRowToken.__handle_start_table_token,
            TableMarkdownRowToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        return "".join(
            [
                output_html,
                "<tr>",
                ParserHelper.newline_character,
            ]
        )

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:  # sourcery skip: for-append-to-extend, for-index-underscore
        _ = (transform_state, next_token)

        ff = [output_html]
        for _i in range(
            cast(
                TableMarkdownRowToken,
                cast(EndMarkdownToken, next_token).start_markdown_token,
            ).delta
        ):
            ff.append("<td></td>\n")
        ff.extend(("</tr>", ParserHelper.newline_character))
        return "".join(ff)


class TableMarkdownRowItemToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the table header markdown token.
    """

    def __init__(
        self,
        leading_whitespace: str,
        column_alignment: Optional[str],
        line_number: int,
        column_number: int,
    ) -> None:
        self.__leading_whitespace: str = leading_whitespace
        self.__column_alignment = column_alignment
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_table_row_item,
            extra_data=self.__compose_extra_data_field(),
            line_number=line_number,
            column_number=column_number,
            extracted_whitespace=leading_whitespace,
            requires_end_token=True,
        )

    @property
    def leading_whitespace(self) -> str:
        """Get the leading whitespace for this row item's text."""
        return self.__leading_whitespace

    @property
    def column_alignment(self) -> Optional[str]:
        """Get the alignment of this row item's column."""
        return self.__column_alignment

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_table_row_item

    # pylint: enable=protected-access

    def __compose_extra_data_field(self) -> Optional[str]:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        composed_field = MarkdownToken.extra_data_separator.join(
            [
                self.leading_whitespace,
                self.column_alignment if self.column_alignment is not None else "",
            ]
        )
        self._set_extra_data(composed_field)
        return composed_field

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TableMarkdownRowItemToken,
            TableMarkdownRowItemToken.__rehydrate_table_row_item_start,
            TableMarkdownRowItemToken.__rehydrate_table_row_item_end,
        )

    @staticmethod
    def __rehydrate_table_row_item_start(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, current_token, context)
        return cast(TableMarkdownRowItemToken, current_token).leading_whitespace

    @staticmethod
    def __rehydrate_table_row_item_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the link reference definition from the token.
        """
        _ = (previous_token, next_token, context)
        return cast(TableMarkdownRowItemToken, current_token).extracted_whitespace

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TableMarkdownRowItemToken,
            TableMarkdownRowItemToken.__handle_start_table_token,
            TableMarkdownRowItemToken.__handle_end_table_token,
        )

    @staticmethod
    def __handle_start_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        table_token = cast(TableMarkdownRowItemToken, next_token)
        if table_token.column_alignment:
            text_to_add = f'<td align="{table_token.column_alignment}">'
        else:
            text_to_add = "<td>"
        return "".join(
            [
                output_html,
                text_to_add,
            ]
        )

    @staticmethod
    def __handle_end_table_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (next_token, transform_state)

        return "".join([output_html, "</td>", ParserHelper.newline_character])



"""
Module to implement a plugin that looks for fenced code blocks without a language specified.
"""

from typing import cast

from pymarkdown.general.constants import Constants
from pymarkdown.plugin_manager.plugin_details import PluginDetails
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.fenced_code_block_markdown_token import (
    FencedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd040(RulePlugin):
    """
    Class to implement a plugin that looks for fenced code blocks without a language specified.
    """

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="fenced-code-language",
            plugin_id="MD040",
            plugin_enabled_by_default=True,
            plugin_description="Fenced code blocks should have a language specified",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md040.md",
        )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        # print(f">>{token}<<")
        if token.is_fenced_code_block:
            fenced_token = cast(FencedCodeBlockMarkdownToken, token)
            # print(f":::>>{fenced_token.extracted_text}<<")
            if not fenced_token.extracted_text.strip(Constants.ascii_whitespace):
                self.report_next_token_error(context, token)



"""
Module to provide for a transformation from tokens to a markdown document.
"""

import collections
import inspect
import logging
from typing import Dict, List, Optional, Tuple, cast

from pymarkdown.extensions.extension_token_types import ExtensionTokenTypes
from pymarkdown.extensions.pragma_token import PragmaToken
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.container_markdown_token import ContainerMarkdownToken
from pymarkdown.tokens.end_of_stream_token import SpecialMarkdownToken
from pymarkdown.tokens.inline_markdown_token import InlineMarkdownToken
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import (
    EndMarkdownToken,
    MarkdownToken,
    MarkdownTokenClass,
)
from pymarkdown.tokens.new_list_item_markdown_token import NewListItemMarkdownToken
from pymarkdown.tokens.ordered_list_start_markdown_token import (
    OrderedListStartMarkdownToken,
)
from pymarkdown.tokens.token_types import TokenTypes
from pymarkdown.tokens.unordered_list_start_markdown_token import (
    UnorderedListStartMarkdownToken,
)
from pymarkdown.transform_markdown.markdown_transform_context import (
    EndMarkdownContainerTokenTransformProtocol,
    EndMarkdownTokenTransformProtocol,
    MarkdownTransformContext,
    StartMarkdownContainerTokenTransformProtocol,
    StartMarkdownTokenTransformProtocol,
)
from pymarkdown.transform_markdown.transform_block_quote import TransformBlockQuote
from pymarkdown.transform_markdown.transform_containers import (
    MarkdownChangeRecord,
    TransformContainers,
)
from pymarkdown.transform_markdown.transform_list_block import TransformListBlock
from pymarkdown.transform_markdown.transform_new_list_item import TransformNewListItem

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-many-lines
class TransformToMarkdown:
    """
    Class to provide for a transformation from tokens to a markdown document.
    """

    def __init__(self) -> None:
        """
        Initializes a new instance of the TransformToMarkdown class.
        """

        self.context = MarkdownTransformContext()

        self.start_container_token_handlers: Dict[
            str,
            StartMarkdownContainerTokenTransformProtocol,
        ] = {}
        self.end_container_token_handlers: Dict[
            str,
            EndMarkdownContainerTokenTransformProtocol,
        ] = {}
        self.start_token_handlers: Dict[
            str,
            StartMarkdownTokenTransformProtocol,
        ] = {}
        self.end_token_handlers: Dict[
            str,
            EndMarkdownTokenTransformProtocol,
        ] = {}

        self.register_container_handlers(
            OrderedListStartMarkdownToken,
            TransformListBlock.rehydrate_list_start,
            TransformListBlock.rehydrate_list_start_end,
        )
        self.register_container_handlers(
            UnorderedListStartMarkdownToken,
            TransformListBlock.rehydrate_list_start,
            TransformListBlock.rehydrate_list_start_end,
        )
        self.register_container_handlers(
            NewListItemMarkdownToken, TransformNewListItem.rehydrate_next_list_item
        )
        self.register_container_handlers(
            BlockQuoteMarkdownToken,
            TransformBlockQuote.rehydrate_block_quote,
            TransformBlockQuote.rehydrate_block_quote_end,
        )

        ert = TokenTypes.get_inline_token_types()
        ert.extend(TokenTypes.get_leaf_token_types())
        ert.extend(TokenTypes.get_special_token_types())
        ert.extend(ExtensionTokenTypes.get_token_types())
        for token_type in ert:
            token_init_fn = token_type.__dict__["__init__"]
            init_parameters = {
                i: "" for i in inspect.getfullargspec(token_init_fn)[0] if i != "self"
            }
            handler_instance = token_type(**init_parameters)
            handler_instance.register_for_markdown_transform(self.register_handlers)

    def __get_token_type_info(self, token_type: type) -> Tuple[str, MarkdownTokenClass]:
        current_token_type = token_type
        assert (
            "get_markdown_token_type" in token_type.__dict__
        ), "Must be present in dictionary."
        token_name = token_type.__dict__["get_markdown_token_type"].__func__()

        while current_token_type not in [
            ContainerMarkdownToken,
            LeafMarkdownToken,
            InlineMarkdownToken,
            SpecialMarkdownToken,
        ]:
            new_bases = list(current_token_type.__bases__)
            assert len(new_bases) == 1, "Only one."
            current_token_type = new_bases[0]

        if current_token_type == ContainerMarkdownToken:
            token_class = MarkdownTokenClass.CONTAINER_BLOCK
        elif current_token_type == LeafMarkdownToken:
            token_class = MarkdownTokenClass.LEAF_BLOCK
        elif current_token_type == InlineMarkdownToken:
            token_class = MarkdownTokenClass.INLINE_BLOCK
        else:
            assert current_token_type == SpecialMarkdownToken, "Default is special."
            token_class = MarkdownTokenClass.SPECIAL

        assert (
            token_name is not None and token_class is not None
        ), "Both must be defined."
        return token_name, token_class

    def register_handlers(
        self,
        token_type: type,
        start_token_handler: StartMarkdownTokenTransformProtocol,
        end_token_handler: Optional[EndMarkdownTokenTransformProtocol] = None,
    ) -> None:
        """
        Register the handlers necessary to deal with token's start and end.
        """
        type_name, type_class = self.__get_token_type_info(token_type)

        assert type_class in (
            MarkdownTokenClass.LEAF_BLOCK,
            MarkdownTokenClass.INLINE_BLOCK,
            MarkdownTokenClass.SPECIAL,
        ), "Must be one of three classes."

        self.start_token_handlers[type_name] = start_token_handler
        if end_token_handler:
            self.end_token_handlers[type_name] = end_token_handler

    def register_container_handlers(
        self,
        token_type: type,
        start_token_handler: StartMarkdownContainerTokenTransformProtocol,
        end_token_handler: Optional[EndMarkdownContainerTokenTransformProtocol] = None,
    ) -> None:
        """
        Register the handlers necessary to deal with token's start and end.
        """
        type_name, type_class = self.__get_token_type_info(token_type)
        assert type_class in [
            MarkdownTokenClass.CONTAINER_BLOCK
        ], "Must be of class CONTAINER_BLOCK"

        self.start_container_token_handlers[type_name] = start_token_handler
        if end_token_handler:
            self.end_container_token_handlers[type_name] = end_token_handler

    def transform(self, actual_tokens: List[MarkdownToken]) -> str:  # noqa: C901
        """
        Transform the incoming token stream back into Markdown.
        """
        container_stack: List[MarkdownToken] = []
        container_records: List[MarkdownChangeRecord] = []
        (
            transformed_data,
            previous_token,
            pragma_token,
        ) = ("", None, None)

        POGGER.debug("---\nTransformToMarkdown\n---")

        for token_index, current_token in enumerate(actual_tokens):
            if current_token.is_list_start:
                list_token = cast(ListStartMarkdownToken, current_token)
                list_token.reset_last_list_item()

            next_token = (
                actual_tokens[token_index + 1]
                if token_index < len(actual_tokens) - 1
                else None
            )

            POGGER.debug(
                f"pre-h>current_token>:{ParserHelper.make_value_visible(current_token)}:"
            )
            (
                new_data,
                pragma_token,
            ) = self.__process_next_token(
                current_token,
                previous_token,
                next_token,
                transformed_data,
                actual_tokens,
                token_index,
            )

            POGGER.debug(
                f"post-h>new_data>:{ParserHelper.make_value_visible(new_data)}:"
            )
            transformed_data_length_before_add = len(transformed_data)
            POGGER.debug(
                f"post-h>transformed_data>:{ParserHelper.make_value_visible(transformed_data)}:"
            )
            transformed_data += new_data
            POGGER.debug(
                f"post-h>transformed_data>:{ParserHelper.make_value_visible(transformed_data)}:"
            )

            transformed_data = TransformContainers.handle_current_token(
                current_token,
                transformed_data,
                container_stack,
                container_records,
                transformed_data_length_before_add,
                actual_tokens,
            )

            POGGER.debug("---")
            previous_token = current_token

        transformed_data = self.__correct_for_final_newline(
            transformed_data, actual_tokens
        )
        transformed_data = (
            transformed_data.replace(ParserLogger.start_range_sequence, "")
            .replace(ParserLogger.end_range_sequence, "")
            .replace(ParserLogger.blah_sequence, "")
        )
        if pragma_token:
            transformed_data = self.__handle_pragma_processing(
                pragma_token, transformed_data
            )

        assert not self.context.block_stack, "Nothing must be left at the end."
        assert (
            not self.context.container_token_stack
        ), "Nothing must be left at the end."
        return transformed_data

    @classmethod
    def __correct_for_final_newline(
        cls, transformed_data: str, actual_tokens: List[MarkdownToken]
    ) -> str:
        was_forced_fenced_end = False
        if actual_tokens[-1].is_fenced_code_block_end:
            prev_end_token = cast(EndMarkdownToken, actual_tokens[-1])
            was_forced_fenced_end = prev_end_token.was_forced
        if (
            transformed_data
            and transformed_data[-1] == ParserHelper.newline_character
            and not (
                was_forced_fenced_end and not actual_tokens[-2].is_fenced_code_block
            )
        ):
            transformed_data = transformed_data[:-1]
        return transformed_data

    @classmethod
    def __handle_pragma_processing(
        cls, pragma_token: PragmaToken, transformed_data: str
    ) -> str:
        ordered_lines = collections.OrderedDict(
            sorted(pragma_token.pragma_lines.items(), key=lambda x: abs(x[0]))
        )

        for next_line_number in ordered_lines:
            POGGER.debug(
                f"pragma-->{ParserHelper.make_value_visible(ordered_lines[next_line_number])}<--"
            )
            detabified_pragma = TabHelper.detabify_string(
                ordered_lines[next_line_number]
            )
            POGGER.debug(
                f"pragma-->{ParserHelper.make_value_visible(detabified_pragma)}<--"
            )

            abs_line_number = abs(next_line_number)
            if abs_line_number == 1:

                transformed_data = (
                    (
                        f"{detabified_pragma}"
                        + f"{ParserHelper.newline_character}{transformed_data}"
                    )
                    if transformed_data
                    else detabified_pragma
                )
            else:
                nth_index = ParserHelper.find_nth_occurrence(
                    transformed_data,
                    ParserHelper.newline_character,
                    abs_line_number - 1,
                )
                transformed_data = (
                    (
                        f"{transformed_data}{ParserHelper.newline_character}"
                        + f"{detabified_pragma}"
                    )
                    if nth_index == -1
                    else (
                        f"{transformed_data[:nth_index]}{ParserHelper.newline_character}"
                        + f"{detabified_pragma}{transformed_data[nth_index:]}"
                    )
                )
        return transformed_data

    # pylint: disable=too-many-arguments
    def __process_next_token(
        self,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
        transformed_data: str,
        actual_tokens: List[MarkdownToken],
        token_index: int,
    ) -> Tuple[str, Optional[PragmaToken]]:
        pragma_token: Optional[PragmaToken] = None
        if current_token.token_name in self.start_container_token_handlers:
            new_data = self.start_container_token_handlers[current_token.token_name](
                self.context,
                current_token,
                previous_token,
                next_token,
                transformed_data,
            )

        elif current_token.token_name in self.start_token_handlers:
            new_data = self.start_token_handlers[current_token.token_name](
                self.context, current_token, previous_token
            )

        elif current_token.is_pragma:
            new_data = ""
            pragma_token = cast(PragmaToken, current_token)
        elif current_token.is_end_token:
            current_end_token = cast(EndMarkdownToken, current_token)
            if current_end_token.type_name in self.end_token_handlers:
                end_handler_fn: EndMarkdownTokenTransformProtocol = (
                    self.end_token_handlers[current_end_token.type_name]
                )
                new_data = end_handler_fn(
                    self.context, current_end_token, previous_token, next_token
                )
            elif current_end_token.type_name in self.end_container_token_handlers:
                end_container_handler_fn: EndMarkdownContainerTokenTransformProtocol = (
                    self.end_container_token_handlers[current_end_token.type_name]
                )
                new_data = end_container_handler_fn(
                    self.context, current_end_token, actual_tokens, token_index
                )
            else:
                raise AssertionError(
                    f"end_current_token>>{current_end_token.type_name}"
                )
        else:
            raise AssertionError(f"current_token>>{current_token}")
        return new_data, pragma_token

    # pylint: enable=too-many-arguments



"""
Module to implement a plugin that ensures that Block Quote elements are surrounded by Blank Lines.
"""

from typing import List, cast

from pymarkdown.plugin_manager.plugin_details import PluginDetails
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.blank_line_markdown_token import BlankLineMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd028(RulePlugin):
    """
    Class to implement a plugin that ensures that Block Quote elements are surrounded by Blank Lines.
    """

    __look_for_end_of_block_quote = 0
    __look_for_blank_lines = 1
    __look_for_start_of_block_quote = 2

    def __init__(self) -> None:
        super().__init__()
        self.__current_state = 0
        self.__found_blank_lines: List[BlankLineMarkdownToken] = []

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-blanks-blockquote",
            plugin_id="MD028",
            plugin_enabled_by_default=True,
            plugin_description="Blank line inside blockquote",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md028.md",
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__current_state = RuleMd028.__look_for_end_of_block_quote
        self.__found_blank_lines = []

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        # print(str(self.__current_state) + "--" + str(token).replace(ParserHelper.newline_character, "\\n"))
        if self.__current_state == RuleMd028.__look_for_end_of_block_quote:
            if token.is_block_quote_end:
                self.__current_state = RuleMd028.__look_for_blank_lines
                self.__found_blank_lines = []
        elif self.__current_state == RuleMd028.__look_for_blank_lines:
            if token.is_blank_line:
                blank_line_token = cast(BlankLineMarkdownToken, token)
                self.__current_state = RuleMd028.__look_for_start_of_block_quote
                self.__found_blank_lines.append(blank_line_token)
            elif not token.is_block_quote_end:
                self.__current_state = RuleMd028.__look_for_end_of_block_quote
        else:
            assert self.__current_state == RuleMd028.__look_for_start_of_block_quote
            if token.is_block_quote_start:
                for next_blank_lines in self.__found_blank_lines:
                    self.report_next_token_error(context, next_blank_lines)
                self.__current_state = RuleMd028.__look_for_end_of_block_quote
            elif token.is_blank_line:
                blank_line_token = cast(BlankLineMarkdownToken, token)
                self.__found_blank_lines.append(blank_line_token)
            else:
                self.__current_state = RuleMd028.__look_for_end_of_block_quote



"""
Module to provide for a container element that can be added to markdown parsing stream.
"""

import logging
from typing import Optional

from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.markdown_token import MarkdownToken, MarkdownTokenClass

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-many-arguments
class ContainerMarkdownToken(MarkdownToken):
    """
    Class to provide for a container element that can be added to markdown parsing stream.
    """

    def __init__(
        self,
        token_name: str,
        extra_data: str,
        line_number: int = 0,
        column_number: int = 0,
        position_marker: Optional[PositionMarker] = None,
    ) -> None:
        MarkdownToken.__init__(
            self,
            token_name,
            MarkdownTokenClass.CONTAINER_BLOCK,
            extra_data,
            line_number=line_number,
            column_number=column_number,
            position_marker=position_marker,
            requires_end_token=True,
        )


# pylint: enable=too-many-arguments



"""
Module to provide for an element that can be added to the stack.
"""

from typing import Any, List, Optional

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.container_markdown_token import ContainerMarkdownToken
from pymarkdown.tokens.fenced_code_block_markdown_token import (
    FencedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.html_block_markdown_token import HtmlBlockMarkdownToken
from pymarkdown.tokens.indented_code_block_markdown_token import (
    IndentedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.new_list_item_markdown_token import NewListItemMarkdownToken
from pymarkdown.tokens.ordered_list_start_markdown_token import (
    OrderedListStartMarkdownToken,
)
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.unordered_list_start_markdown_token import (
    UnorderedListStartMarkdownToken,
)


class StackToken:
    """
    Class to provide for an element to place on the stack.
    """

    _stack_base_document = "document"
    _stack_unordered_list = "ulist"
    _stack_ordered_list = "olist"
    _stack_html_block = "html-block"
    _stack_block_quote = "block-quote"
    _stack_fenced_code = "fcode-block"
    _stack_indented_code = "icode-block"
    _stack_paragraph = "para"
    _stack_link_definition = "linkdef"
    _stack_table_rows = "table-rows"

    def __init__(
        self,
        type_name: str,
        matching_markdown_token: Optional[MarkdownToken] = None,
        extra_data: Optional[str] = None,
    ) -> None:
        self.__type_name, self.__extra_data, self.__matching_markdown_token = (
            type_name,
            extra_data,
            matching_markdown_token,
        )

    def __str__(self) -> str:
        add_extra = f":{self.extra_data}" if self.extra_data else ""
        return f"StackToken({self.type_name}{add_extra})"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: Any) -> bool:
        """
        Overrides the default implementation
        """
        assert isinstance(other, StackToken), "Cover other objects."
        return self.type_name == other.type_name and self.extra_data == other.extra_data

    @property
    def type_name(self) -> str:
        """
        Returns the type name associated with this stack token.
        """
        return self.__type_name

    @property
    def extra_data(self) -> Optional[str]:
        """
        Returns the extra data associated with this stack token.
        """
        return self.__extra_data

    @property
    def matching_markdown_token(self) -> Optional[MarkdownToken]:
        """
        Returns the matching markdown associated with this stack token.
        """
        return self.__matching_markdown_token

    def reset_matching_markdown_token(
        self, new_matching_markdown_token: MarkdownToken
    ) -> None:
        """
        Reset the matching markdown token.  To be used only when rewinding.
        """
        self.__matching_markdown_token = new_matching_markdown_token

    def generate_close_markdown_token_from_stack_token(
        self,
        extracted_whitespace: str = "",
        extra_end_data: Optional[str] = None,
        was_forced: bool = False,
    ) -> EndMarkdownToken:
        """
        Generate the token emitted to close off the current stack token
        """

        assert (
            self._stack_link_definition != self.type_name
        ), "A stack LRD token must never create a matching close token."
        assert self.matching_markdown_token, str(self)

        return EndMarkdownToken(
            self.type_name,
            extracted_whitespace,
            extra_end_data,
            self.matching_markdown_token,
            was_forced,
        )

    @property
    def is_document(self) -> bool:
        """
        Is this stack token a document token?
        """
        return self.type_name == self._stack_base_document

    @property
    def is_unordered_list(self) -> bool:
        """
        Is this stack token the unordered list token?
        """
        return self.type_name == self._stack_unordered_list

    @property
    def is_ordered_list(self) -> bool:
        """
        Is this stack token the ordered list token?
        """
        return self.type_name == self._stack_ordered_list

    @property
    def is_list(self) -> bool:
        """
        Is this stack token one of the list tokens?
        """
        return self.is_ordered_list or self.is_unordered_list

    @property
    def is_html_block(self) -> bool:
        """
        Is this stack token a html block token?
        """
        return self.type_name == self._stack_html_block

    @property
    def is_code_block(self) -> bool:
        """
        Is this stack token a fenced code block or indented code block token?
        """
        return self.is_fenced_code_block or self.is_indented_code_block

    @property
    def is_fenced_code_block(self) -> bool:
        """
        Is this stack token a fenced code block token?
        """
        return self.type_name == self._stack_fenced_code

    @property
    def is_indented_code_block(self) -> bool:
        """
        Is this stack token an indented code block token?
        """
        return self.type_name == self._stack_indented_code

    @property
    def is_block_quote(self) -> bool:
        """
        Is this stack token a block quote token?
        """
        return self.type_name == self._stack_block_quote

    @property
    def is_paragraph(self) -> bool:
        """
        Is this stack token a paragraph token?
        """
        return self.type_name == self._stack_paragraph

    @property
    def was_link_definition_started(self) -> bool:
        """
        Is this stack token a link definition started token?
        """
        return self.type_name == self._stack_link_definition

    @property
    def was_table_block_started(self) -> bool:
        """
        Is this stack token a table block token?
        """
        return self.type_name == self._stack_table_rows


class DocumentStackToken(StackToken):
    """
    Class to provide for a stack token at the root.
    """

    def __init__(self) -> None:
        StackToken.__init__(self, StackToken._stack_base_document)


class ParagraphStackToken(StackToken):
    """
    Class to provide for a stack token for a paragraph.
    """

    def __init__(self, matching_markdown_token: ParagraphMarkdownToken) -> None:
        StackToken.__init__(
            self,
            StackToken._stack_paragraph,
            matching_markdown_token=matching_markdown_token,
        )


class BlockQuoteStackToken(StackToken):
    """
    Class to provide for a stack token for a block quote.
    """

    def __init__(self, matching_markdown_token: BlockQuoteMarkdownToken) -> None:
        StackToken.__init__(
            self,
            StackToken._stack_block_quote,
            matching_markdown_token=matching_markdown_token,
        )


class IndentedCodeBlockStackToken(StackToken):
    """
    Class to provide for a stack token for an indented code block.
    """

    def __init__(self, matching_markdown_token: IndentedCodeBlockMarkdownToken) -> None:
        StackToken.__init__(
            self,
            StackToken._stack_indented_code,
            matching_markdown_token=matching_markdown_token,
        )


class FencedCodeBlockStackToken(StackToken):
    """
    Class to provide for a stack token for a fenced code block.
    """

    def __init__(
        self,
        code_fence_character: str,
        fence_character_count: int,
        whitespace_start_count: int,
        matching_markdown_token: FencedCodeBlockMarkdownToken,
    ) -> None:
        (
            self.__code_fence_character,
            self.__fence_character_count,
            self.__whitespace_start_count,
        ) = (code_fence_character, fence_character_count, whitespace_start_count)
        StackToken.__init__(
            self,
            StackToken._stack_fenced_code,
            matching_markdown_token=matching_markdown_token,
            extra_data=":".join(
                [
                    self.__code_fence_character,
                    str(self.__fence_character_count),
                    str(self.__whitespace_start_count),
                ]
            ),
        )

    @property
    def code_fence_character(self) -> str:
        """
        Returns the fence character associated with this stack token.
        """
        return self.__code_fence_character

    @property
    def fence_character_count(self) -> int:
        """
        Returns the fence character count associated with this stack token.
        """
        return self.__fence_character_count

    @property
    def whitespace_start_count(self) -> int:
        """
        Returns the count of whitespaces preceding this stack token.
        """
        return self.__whitespace_start_count


class ListStackToken(StackToken):
    """
    Class to provide for a stack token for a list.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        type_name: str,
        indent_level: int,
        list_character: str,
        ws_before_marker: int,
        ws_after_marker: int,
        start_index: int,
        matching_markdown_token: ContainerMarkdownToken,
    ) -> None:
        (
            self.__indent_level,
            self.__list_character,
            self.__ws_before_marker,
            self.__ws_after_marker,
            self.__start_index,
        ) = (
            indent_level,
            list_character,
            ws_before_marker,
            ws_after_marker,
            start_index,
        )
        self.__last_new_list_token: Optional[NewListItemMarkdownToken] = None

        StackToken.__init__(
            self,
            type_name,
            matching_markdown_token=matching_markdown_token,
            extra_data=":".join(
                [
                    str(self.__indent_level),
                    self.__list_character,
                    str(self.__ws_before_marker),
                    str(self.__ws_after_marker),
                    str(self.__start_index),
                ]
            ),
        )

    # pylint: enable=too-many-arguments
    @property
    def indent_level(self) -> int:
        """
        Returns the indent level associated with this stack token.
        """
        return self.__indent_level

    @property
    def list_character(self) -> str:
        """
        Returns the list character associated with this stack token.
        """
        return self.__list_character

    @property
    def ws_before_marker(self) -> int:
        """
        Returns the whitespace occurring before this stack token.
        """
        return self.__ws_before_marker

    @property
    def ws_after_marker(self) -> int:
        """
        Returns the whitespace occurring after this stack token.
        """
        return self.__ws_after_marker

    @property
    def start_index(self) -> int:
        """
        Returns the start index for this stack token.
        """
        return self.__start_index

    @property
    def last_new_list_token(self) -> Optional[NewListItemMarkdownToken]:
        """
        Returns the last new-list token associated with this stack token.
        """
        return self.__last_new_list_token

    def set_last_new_list_token(self, new_list_token: NewListItemMarkdownToken) -> None:
        """
        Set the last new-list token associated with this stack token.
        """
        self.__last_new_list_token = new_list_token


class OrderedListStackToken(ListStackToken):
    """
    Class to provide for a stack token for an ordered list.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        indent_level: int,
        list_character: str,
        ws_before_marker: int,
        ws_after_marker: int,
        start_index: int,
        matching_markdown_token: OrderedListStartMarkdownToken,
    ) -> None:
        ListStackToken.__init__(
            self,
            StackToken._stack_ordered_list,
            indent_level,
            list_character,
            ws_before_marker,
            ws_after_marker,
            start_index,
            matching_markdown_token,
        )

    # pylint: enable=too-many-arguments


class UnorderedListStackToken(ListStackToken):
    """
    Class to provide for a stack token for an unordered list.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        indent_level: int,
        list_character: str,
        ws_before_marker: int,
        ws_after_marker: int,
        start_index: int,
        matching_markdown_token: UnorderedListStartMarkdownToken,
    ) -> None:
        ListStackToken.__init__(
            self,
            StackToken._stack_unordered_list,
            indent_level,
            list_character,
            ws_before_marker,
            ws_after_marker,
            start_index,
            matching_markdown_token,
        )

    # pylint: enable=too-many-arguments


class HtmlBlockStackToken(StackToken):
    """
    Class to provide for a stack token for a html block.
    """

    def __init__(
        self, html_block_type: str, matching_markdown_token: HtmlBlockMarkdownToken
    ) -> None:
        self.__html_block_type = html_block_type
        extra_data = html_block_type
        StackToken.__init__(
            self,
            StackToken._stack_html_block,
            matching_markdown_token=matching_markdown_token,
            extra_data=extra_data,
        )

    @property
    def html_block_type(self) -> str:
        """
        Returns the html block type associated with this stack token.
        """
        return self.__html_block_type


# pylint: disable=too-many-instance-attributes
class LinkDefinitionStackToken(StackToken):
    """
    Class to provide for a stack token for a possible link definition.
    """

    def __init__(
        self, extracted_whitespace: str, position_marker: PositionMarker
    ) -> None:
        (
            self.__extracted_whitespace,
            self.__start_position_marker,
        ) = (
            extracted_whitespace,
            position_marker,
        )
        self.original_stack_depth: Optional[int] = None
        self.original_document_depth: Optional[int] = None
        self.last_block_quote_stack_token: Optional[StackToken] = None
        self.last_block_quote_markdown_token_index: Optional[int] = None
        self.copy_of_last_block_quote_markdown_token: Optional[
            BlockQuoteMarkdownToken
        ] = None
        self.copy_of_token_stack: Optional[List[StackToken]] = None

        self.x1_token: Optional[MarkdownToken] = None
        self.copy_of_x1_token: Optional[MarkdownToken] = None
        self.x1_token_index: Optional[int] = None

        self.__continuation_lines: List[str] = []
        self.__unmodified_lines: List[str] = []
        # self.__original_lines: List[str] = []
        StackToken.__init__(self, StackToken._stack_link_definition)

    @property
    def extracted_whitespace(self) -> Optional[str]:
        """
        Returns the extracted whitespace associated with this stack token.
        """
        return self.__extracted_whitespace

    @property
    def continuation_lines(self) -> List[str]:
        """
        Returns the continuation lines associated with this stack token.
        """
        return self.__continuation_lines

    @property
    def unmodified_lines(self) -> List[str]:
        """
        Returns the unmodified continuation lines associated with this stack token.
        """
        return self.__unmodified_lines

    # @property
    # def original_lines(self) -> List[str]:
    #     """
    #     Returns the original lines associated with this stack token.
    #     """
    #     return self.__original_lines

    @property
    def start_position_marker(self) -> PositionMarker:
        """
        Returns the start position associated with this stack token.
        """
        return self.__start_position_marker

    def add_continuation_line(self, new_line: str) -> None:
        """
        Add the line to the collection of lines to keep as "continuations".
        """
        self.__continuation_lines.append(new_line)

    def add_unmodified_line(self, new_line: str) -> None:
        """
        Add the line to the collection of lines to keep as "umodified line".
        These are the same as the continuation_lines values, just with the exact
        text that was presented to the parser.
        """
        self.__unmodified_lines.append(new_line)

    # def add_original_line(self, new_line: str) -> None:
    #     """
    #     ...
    #     """
    #     self.__original_lines.append(new_line)

    def add_joined_lines_before_suffix(self, join_suffix: str) -> str:
        """
        Grab the continuation lines as a single line.
        """

        return (
            f"{ParserHelper.newline_character.join(self.continuation_lines)}"
            + f"{ParserHelper.newline_character}{join_suffix}"
            if self.continuation_lines
            else join_suffix
        )


# pylint: enable=too-many-instance-attributes


class TableBlockStackToken(StackToken):
    """
    Class to provide for a stack token for a possible table block.
    """

    def __init__(
        self, extracted_whitespace: str, position_marker: PositionMarker
    ) -> None:
        (self.__start_position_marker, self.__extracted_whitespace) = (
            position_marker,
            extracted_whitespace,
        )
        self.original_stack_depth: Optional[int] = None
        self.original_document_depth: Optional[int] = None
        self.last_block_quote_stack_token: Optional[StackToken] = None
        self.last_block_quote_markdown_token_index: Optional[int] = None
        self.copy_of_last_block_quote_markdown_token: Optional[
            BlockQuoteMarkdownToken
        ] = None
        self.copy_of_token_stack: Optional[List[StackToken]] = None

        self.x1_token: Optional[MarkdownToken] = None
        self.copy_of_x1_token: Optional[MarkdownToken] = None
        self.x1_token_index: Optional[int] = None

        self.__continuation_lines: List[str] = []
        self.__unmodified_lines: List[str] = []
        StackToken.__init__(self, StackToken._stack_table_rows)

    @property
    def extracted_whitespace(self) -> Optional[str]:
        """
        Returns the extracted whitespace associated with this stack token.
        """
        return self.__extracted_whitespace

    @property
    def continuation_lines(self) -> List[str]:
        """
        Returns the continuation lines associated with this stack token.
        """
        return self.__continuation_lines

    @property
    def unmodified_lines(self) -> List[str]:
        """
        Returns the unmodified continuation lines associated with this stack token.
        """
        return self.__unmodified_lines

    @property
    def start_position_marker(self) -> PositionMarker:
        """
        Returns the start position associated with this stack token.
        """
        return self.__start_position_marker

    def add_continuation_line(self, new_line: str) -> None:
        """
        Add the line to the collection of lines to keep as "continuations".
        """
        self.__continuation_lines.append(new_line)

    def add_unmodified_line(self, new_line: str) -> None:
        """
        Add the line to the collection of lines to keep as "umodified line".
        These are the same as the continuation_lines values, just with the exact
        text that was presented to the parser.
        """
        self.__unmodified_lines.append(new_line)

    def add_joined_lines_before_suffix(self, join_suffix: str) -> str:
        """
        Grab the continuation lines as a single line.
        """

        return (
            f"{ParserHelper.newline_character.join(self.continuation_lines)}"
            + f"{ParserHelper.newline_character}{join_suffix}"
            if self.continuation_lines
            else join_suffix
        )



"""
Module to allow for a critical error within a plugin to be encapsulated
    and reported.
"""

from typing import Optional

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.tokens.markdown_token import MarkdownToken


class BadPluginError(Exception):
    """
    Class to allow for a critical error within a plugin to be encapsulated
    and reported.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        plugin_id: Optional[str] = None,
        plugin_action: Optional[str] = None,
        file_name: Optional[str] = None,
        class_name: Optional[str] = None,
        field_name: Optional[str] = None,
        is_constructor: bool = False,
        is_empty: bool = False,
        formatted_message: Optional[str] = None,
        line_number: int = 0,
        column_number: int = 0,
        actual_line: Optional[str] = None,
        actual_token: Optional[MarkdownToken] = None,
        cause: Optional[Exception] = None,
    ) -> None:
        if not formatted_message:
            if file_name:
                formatted_message = BadPluginError.__create_file_name_message(
                    file_name, class_name, is_constructor
                )
            elif class_name:
                formatted_message = BadPluginError.__create_class_name_message(
                    class_name, field_name, is_empty
                )
            else:
                formatted_message = BadPluginError.__create_exception_message(
                    cause, plugin_id, plugin_action
                )

            formatted_message = BadPluginError.__add_suffix(
                formatted_message, actual_line, actual_token, line_number, column_number
            )
        super().__init__(formatted_message)

    # pylint: enable=too-many-arguments

    @staticmethod
    def __create_file_name_message(
        file_name: Optional[str], class_name: Optional[str], is_constructor: bool
    ) -> str:
        if class_name:
            return (
                f"Plugin file named '{file_name}' threw an exception in the constructor for the class '{class_name}'."
                if is_constructor
                else f"Plugin file named '{file_name}' does not contain a class named '{class_name}'."
            )

        return f"Plugin file named '{file_name}' cannot be loaded."

    @staticmethod
    def __create_class_name_message(
        class_name: Optional[str], field_name: Optional[str], is_empty: bool
    ) -> str:
        if field_name:
            return (
                f"Plugin class '{class_name}' returned an empty value for field name '{field_name}'."
                if is_empty
                else f"Plugin class '{class_name}' returned an improperly typed value for field name '{field_name}'."
            )
        return f"Plugin class '{class_name}' had a critical failure loading the plugin details."

    @staticmethod
    def __create_exception_message(
        cause: Optional[Exception],
        plugin_id: Optional[str],
        plugin_action: Optional[str],
    ) -> str:
        if cause and isinstance(cause, ValueError):
            return str(cause)
        assert plugin_id
        return f"Plugin id '{plugin_id.upper()}' had a critical failure during the '{plugin_action}' action."

    @staticmethod
    def __add_suffix(
        formatted_message: str,
        actual_line: Optional[str],
        actual_token: Optional[MarkdownToken],
        line_number: int,
        column_number: int,
    ) -> str:
        if line_number:
            position_message = (
                f"({line_number},{column_number})"
                if column_number
                else f"(Line {line_number})"
            )
            formatted_message = f"{position_message}: {formatted_message}"
        if actual_line:
            formatted_message = f"{formatted_message}\nActual Line: {actual_line}"
        if actual_token:
            formatted_message = f"{formatted_message}\nActual Token: {ParserHelper.make_value_visible(actual_token)}"
        return formatted_message



"""
Module to contain lists of the available token types.
"""

from typing import List

from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.blank_line_markdown_token import BlankLineMarkdownToken
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.email_autolink_markdown_token import EmailAutolinkMarkdownToken
from pymarkdown.tokens.emphasis_markdown_token import EmphasisMarkdownToken
from pymarkdown.tokens.end_of_stream_token import EndOfStreamToken
from pymarkdown.tokens.fenced_code_block_markdown_token import (
    FencedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.hard_break_markdown_token import HardBreakMarkdownToken
from pymarkdown.tokens.html_block_markdown_token import HtmlBlockMarkdownToken
from pymarkdown.tokens.image_start_markdown_token import ImageStartMarkdownToken
from pymarkdown.tokens.indented_code_block_markdown_token import (
    IndentedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.inline_code_span_markdown_token import (
    InlineCodeSpanMarkdownToken,
)
from pymarkdown.tokens.link_reference_definition_markdown_token import (
    LinkReferenceDefinitionMarkdownToken,
)
from pymarkdown.tokens.link_start_markdown_token import LinkStartMarkdownToken
from pymarkdown.tokens.new_list_item_markdown_token import NewListItemMarkdownToken
from pymarkdown.tokens.ordered_list_start_markdown_token import (
    OrderedListStartMarkdownToken,
)
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.setext_heading_markdown_token import SetextHeadingMarkdownToken
from pymarkdown.tokens.table_markdown_tokens import (
    TableMarkdownBodyToken,
    TableMarkdownHeaderItemToken,
    TableMarkdownHeaderToken,
    TableMarkdownRowItemToken,
    TableMarkdownRowToken,
    TableMarkdownToken,
)
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken
from pymarkdown.tokens.thematic_break_markdown_token import ThematicBreakMarkdownToken
from pymarkdown.tokens.unordered_list_start_markdown_token import (
    UnorderedListStartMarkdownToken,
)
from pymarkdown.tokens.uri_autolink_markdown_token import UriAutolinkMarkdownToken


class TokenTypes:
    """
    Class to contain lists of the available token types.
    """

    __SPECIAL_TOKEN_TYPES: List[type] = [
        EndOfStreamToken,
    ]

    __INLINE_TOKEN_TYPES: List[type] = [
        EmphasisMarkdownToken,
        UriAutolinkMarkdownToken,
        EmailAutolinkMarkdownToken,
        RawHtmlMarkdownToken,
        HardBreakMarkdownToken,
        InlineCodeSpanMarkdownToken,
        ImageStartMarkdownToken,
        LinkStartMarkdownToken,
        TextMarkdownToken,
    ]
    __LEAF_TOKEN_TYPES: List[type] = [
        HtmlBlockMarkdownToken,
        ThematicBreakMarkdownToken,
        LinkReferenceDefinitionMarkdownToken,
        BlankLineMarkdownToken,
        SetextHeadingMarkdownToken,
        AtxHeadingMarkdownToken,
        FencedCodeBlockMarkdownToken,
        IndentedCodeBlockMarkdownToken,
        ParagraphMarkdownToken,
        TableMarkdownToken,
        TableMarkdownHeaderToken,
        TableMarkdownHeaderItemToken,
        TableMarkdownBodyToken,
        TableMarkdownRowToken,
        TableMarkdownRowItemToken,
    ]

    __CONTAINER_TOKEN_TYPES: List[type] = [
        BlockQuoteMarkdownToken,
        NewListItemMarkdownToken,
        UnorderedListStartMarkdownToken,
        OrderedListStartMarkdownToken,
    ]

    @staticmethod
    def get_special_token_types() -> List[type]:
        """
        Get a list of all available inline token types.
        """
        return TokenTypes.__SPECIAL_TOKEN_TYPES[:]

    @staticmethod
    def get_inline_token_types() -> List[type]:
        """
        Get a list of all available inline token types.
        """
        return TokenTypes.__INLINE_TOKEN_TYPES[:]

    @staticmethod
    def get_leaf_token_types() -> List[type]:
        """
        Get a list of all available leaf token types.
        """
        return TokenTypes.__LEAF_TOKEN_TYPES[:]

    @staticmethod
    def get_container_token_types() -> List[type]:
        """
        Get a list of all available container token types.
        """
        return TokenTypes.__CONTAINER_TOKEN_TYPES[:]



"""
Module to provide for an encapsulation of a generic list start element.
"""

# pylint: disable=too-many-instance-attributes
import logging
from typing import Optional, Union

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.container_markdown_token import ContainerMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.new_list_item_markdown_token import NewListItemMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class ListStartMarkdownToken(ContainerMarkdownToken):
    """
    Class to provide for an encapsulation of a generic list start element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        token_name: str,
        position_marker: PositionMarker,
        list_start_sequence: str,
        list_start_content: str,
        indent_level: int,
        tabbed_adjust: int,
        extracted_whitespace: str,
        tabbed_whitespace_to_add: Optional[str],
    ) -> None:
        ContainerMarkdownToken.__init__(
            self,
            token_name,
            "",
            position_marker=position_marker,
        )
        POGGER.debug("tabbed_adjust>:$:<", tabbed_adjust)
        POGGER.debug("type(tabbed_adjust)>:$:<", type(tabbed_adjust))
        self.leading_spaces_index = 0
        self.is_loose = True
        (
            self.__list_start_sequence,
            self.__list_start_content,
            self.__indent_level,
            self.__extracted_whitespace,
            self.__tabbed_whitespace_to_add,
            self.__tabbed_adjust,
        ) = (
            list_start_sequence,
            list_start_content,
            indent_level,
            extracted_whitespace,
            tabbed_whitespace_to_add,
            tabbed_adjust,
        )
        self.__last_new_list_token: Optional[NewListItemMarkdownToken] = None
        self.__leading_spaces: Optional[str] = None
        self.__compose_extra_data_field()

    # pylint: enable=too-many-arguments

    @property
    def list_start_sequence(self) -> str:
        """
        Returns the sequence used to start this list element.
        """
        return self.__list_start_sequence

    @property
    def list_start_content(self) -> str:
        """
        Returns the content used to start this list element.
        """
        return self.__list_start_content

    @property
    def indent_level(self) -> int:
        """
        Returns the indent level to apply for the list element.
        """
        return self.__indent_level

    @property
    def extracted_whitespace(self) -> str:
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def tabbed_extracted_whitespace(self) -> Optional[str]:
        """
        extracted_whitespace, but with any tabs.
        """
        return self.__tabbed_whitespace_to_add

    @property
    def tabbed_adjust(self) -> int:
        """
        Adjustments for tabbed.
        """
        return self.__tabbed_adjust

    @property
    def leading_spaces(self) -> Optional[str]:
        """
        Returns the leading spaces that occur before the list element.
        """
        return self.__leading_spaces

    def reset_last_list_item(self) -> None:
        """
        Reset the last new list token seens for this list.
        """
        self.__last_new_list_token = None

    def adjust_for_new_list_item(
        self,
        new_list_item_token: NewListItemMarkdownToken,
        skip_adjustment: bool = False,
    ) -> None:
        """
        Adjust this token for a new list item (uses copy to keep track).
        """
        # assert new_list_item_token and new_list_item_token.is_new_list_item
        self.__last_new_list_token = new_list_item_token

        if not skip_adjustment:
            self.__indent_level, self.__extracted_whitespace = (
                new_list_item_token.indent_level,
                new_list_item_token.extracted_whitespace,
            )

    def __compose_extra_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """

        item_list = [
            self.__list_start_sequence,
            self.__list_start_content,
            str(self.__indent_level),
            self.__extracted_whitespace,
        ]
        is_leading_spaces_valid = self.__leading_spaces is not None
        is_tabbed_whitespace_valid = self.__tabbed_whitespace_to_add is not None
        is_tabbed_adjust_valid = (
            isinstance(self.__tabbed_adjust, int) and self.__tabbed_adjust >= 0
        )

        if (
            is_leading_spaces_valid
            or is_tabbed_whitespace_valid
            or is_tabbed_adjust_valid
        ):
            item_list.append(self.__leading_spaces or "")
        if is_tabbed_whitespace_valid or is_tabbed_adjust_valid:
            item_list.append(self.__tabbed_whitespace_to_add or "")
        if is_tabbed_adjust_valid:
            item_list.append(str(self.__tabbed_adjust))
        self._set_extra_data(MarkdownToken.extra_data_separator.join(item_list))

    def remove_last_leading_space(self) -> Optional[str]:
        """
        Remove the last leading space and return it.
        """
        assert (
            self.__leading_spaces is not None
        ), "Leading spaces must be defined by now."
        POGGER.debug("__leading_spaces>>:$:<<", self.__leading_spaces)
        last_separator_index = self.__leading_spaces.rfind("\n")
        if last_separator_index == -1:
            extracted_text = self.__leading_spaces
            self.__leading_spaces = None
        else:
            extracted_text = self.__leading_spaces[last_separator_index + 1 :]
            self.__leading_spaces = self.__leading_spaces[:last_separator_index]
        self.__compose_extra_data_field()
        POGGER.debug("__leading_spaces>>:$:<<", self.__leading_spaces)
        return extracted_text

    def add_leading_spaces(self, ws_add: str) -> None:
        """
        Add any leading spaces to the token, separating them with line feeds.
        """
        POGGER.debug("__leading_spaces>>:$:<<", self.__leading_spaces)
        POGGER.debug("add_leading_spaces>>:$:<<", ws_add)
        self.__leading_spaces = (
            ws_add
            if self.__leading_spaces is None
            else f"{self.__leading_spaces}{ParserHelper.newline_character}{ws_add}"
        )
        POGGER.debug("__leading_spaces>>:$:<<", self.__leading_spaces)
        self.__compose_extra_data_field()

    @property
    def last_new_list_token(self) -> Optional[NewListItemMarkdownToken]:
        """
        Returns the last new-list token associated with this stack token.
        """
        return self.__last_new_list_token

    def set_extracted_whitespace(self, new_whitespace: str) -> None:
        """
        Set the extracted whitespace for the token.  To be used sparingly.
        """
        self.__extracted_whitespace = new_whitespace
        self.__compose_extra_data_field()

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "list_start_content" and isinstance(field_value, str):
            self.__list_start_content = field_value
            self.__compose_extra_data_field()
            return True
        if field_name == "list_start_sequence" and isinstance(field_value, str):
            self.__list_start_sequence = field_value
            self.__compose_extra_data_field()
            return True
        if field_name == "extracted_whitespace" and isinstance(field_value, str):
            self.__extracted_whitespace = field_value
            self.__compose_extra_data_field()
            return True
        if field_name == "indent_level" and isinstance(field_value, int):
            self.__indent_level = field_value
            self.__compose_extra_data_field()
            return True
        if field_name == "leading_spaces" and isinstance(field_value, str):
            self.__leading_spaces = field_value
            self.__compose_extra_data_field()
            return True
        return super()._modify_token(field_name, field_value)


# pylint: enable=too-many-instance-attributes



"""
Module to work with the rules to keep track of the current container "leading space" index.
"""

from dataclasses import dataclass
from typing import List, Tuple, cast

from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.setext_heading_markdown_token import SetextHeadingMarkdownToken


@dataclass
class ClosedContainerAdjustments:
    """
    Keep track of line space used by already closed containers.
    """

    adjustment: int = 0
    count: int = 0
    count2: int = 0
    list_nudge_count: int = 0


class LeadingSpaceIndexTracker:
    """
    Class to track the leading spaces for each container.
    """

    def __init__(self) -> None:
        self.__closed_container_adjustments: List[ClosedContainerAdjustments] = []
        self.__container_token_stack: List[MarkdownToken] = []
        self.__end_tokens: List[EndMarkdownToken] = []
        self.__real_end_tokens: List[EndMarkdownToken] = []
        self.__since_last_non_end_token: List[MarkdownToken] = []

    def clear(self) -> None:
        """
        Clear all tracking (at the beginning of a new file).
        """
        self.__closed_container_adjustments.clear()
        self.__container_token_stack.clear()
        self.__end_tokens.clear()
        self.__real_end_tokens.clear()

    def open_container(self, token: MarkdownToken) -> None:
        """
        Open a new container.
        """
        assert token.is_block_quote_start or token.is_list_start
        self.__container_token_stack.append(token)
        self.__closed_container_adjustments.append(ClosedContainerAdjustments())

    def nudge_list_container(
        self, token: MarkdownToken, nudge_positive: bool = True
    ) -> None:
        if nudge_positive:
            assert token.is_new_list_item
            assert (
                self.__container_token_stack
                and self.__container_token_stack[-1].is_list_start
            )
            self.__closed_container_adjustments[-1].list_nudge_count += 1
        else:
            self.__closed_container_adjustments[-1].list_nudge_count -= 1

    def register_container_end(self, token: MarkdownToken) -> None:
        """
        Register the end of the current container for processing once the
        next non-end token is encountered.
        """
        assert token.is_block_quote_end or token.is_list_end
        self.__end_tokens.append(cast(EndMarkdownToken, token))
        self.__real_end_tokens.append(cast(EndMarkdownToken, token))

    def have_any_registered_container_ends(self) -> bool:
        """
        Check to see if any container ends have been registered and not processed.
        """
        return bool(self.__end_tokens)

    def process_container_end(
        self, token: MarkdownToken
    ) -> Tuple[MarkdownToken, EndMarkdownToken]:
        """
        Process a registered container end.
        """

        if self.__container_token_stack[-1].is_block_quote_start:
            self.__process_container_end_block_quote(token)
        else:
            self.__process_container_end_list(token)

        del self.__closed_container_adjustments[-1]
        end_token = self.__real_end_tokens[0]
        del self.__real_end_tokens[0]
        del self.__end_tokens[-1]

        last_token_on_stack = self.__container_token_stack[-1]
        del self.__container_token_stack[-1]
        return last_token_on_stack, end_token

    def track_since_last_non_end_token(self, token: MarkdownToken) -> None:
        """
        Keep track of the last non-end token and any end tokens since then.
        """
        if not token.is_end_token:
            self.__since_last_non_end_token.clear()
        self.__since_last_non_end_token.append(token)

    # def since_last_non_end_token(self) -> List[MarkdownToken]:
    #     return self.__since_last_non_end_token[:]

    def get_closed_container_info(self, index: int) -> ClosedContainerAdjustments:
        """
        Get the current information on all closed containers.
        """
        return self.__closed_container_adjustments[index]

    def in_at_least_one_container(self) -> bool:
        """
        Check to see if we are in at least one container.
        """
        return bool(self.__container_token_stack)

    def get_container_stack_size(self) -> int:
        """
        Get the number of containers that we are currently in.
        """
        return len(self.__container_token_stack)

    def get_container_stack_item(self, index: int) -> MarkdownToken:
        """
        Get a specific container from the stack.
        """
        return self.__container_token_stack[index]

    @staticmethod
    def calculate_token_line_number(token: MarkdownToken) -> int:
        """
        Since setext tokens are "weird", helper function to calculate the line number.
        """
        if token.is_setext_heading:
            setext_token = cast(SetextHeadingMarkdownToken, token)
            return setext_token.original_line_number
        return token.line_number

    def __process_container_end_block_quote(self, token: MarkdownToken) -> None:
        for stack_index in range(len(self.__container_token_stack) - 2, -1, -1):
            current_stack_token = self.__container_token_stack[stack_index]
            if current_stack_token.is_block_quote_start:
                line_number_delta = (
                    LeadingSpaceIndexTracker.calculate_token_line_number(token)
                    - self.__container_token_stack[-1].line_number
                )
                if self.__end_tokens[-1].extra_end_data is not None:
                    line_number_delta += 1
                self.__closed_container_adjustments[
                    stack_index
                ].adjustment += line_number_delta
                self.__closed_container_adjustments[stack_index].count += 1

                does_current_container_have_weird_kludge_six = (
                    self.__container_token_stack[-1].is_block_quote_start
                    and cast(
                        BlockQuoteMarkdownToken, self.__container_token_stack[-1]
                    ).weird_kludge_six
                )
                was_last_non_end_token_blank_line = (
                    self.__end_tokens
                    and self.__since_last_non_end_token
                    and self.__since_last_non_end_token[0].is_blank_line
                )

                if (
                    does_current_container_have_weird_kludge_six
                    and not was_last_non_end_token_blank_line
                ):
                    self.__closed_container_adjustments[stack_index].count2 += 1
                break

    def __process_container_end_list(self, token: MarkdownToken) -> None:

        g = 2
        for stack_index in range(len(self.__container_token_stack) - g, -1, -1):
            current_stack_token = self.__container_token_stack[stack_index]
            if current_stack_token.is_list_start:
                line_number_delta = (
                    LeadingSpaceIndexTracker.calculate_token_line_number(token)
                    - self.__container_token_stack[-1].line_number
                )
                if (
                    self.__end_tokens[-1].start_markdown_token.line_number
                    == current_stack_token.line_number
                ):
                    line_number_delta -= 1
                self.__closed_container_adjustments[
                    stack_index
                ].adjustment += line_number_delta
                break

    def get_tokens_block_quote_bleading_space_index(
        self, token: MarkdownToken, alternate_index: int = -1
    ) -> int:
        """
        Get the index of the token within the contain block quote's bleading_spaces string.
        """
        container_index = (
            self.get_container_stack_size() - 1
            if alternate_index < 0
            else alternate_index
        )
        assert self.__container_token_stack[container_index].is_block_quote_start
        block_quote_token = cast(
            BlockQuoteMarkdownToken, self.__container_token_stack[container_index]
        )
        last_closed_container_info = self.__closed_container_adjustments[-1]

        assert (
            block_quote_token.bleading_spaces is not None
        ), "At least one line should have been processed."
        return (
            LeadingSpaceIndexTracker.calculate_token_line_number(token)
            - block_quote_token.line_number
        ) - (last_closed_container_info.adjustment - last_closed_container_info.count2)

    def get_tokens_list_leading_space_index(
        self, token: MarkdownToken, alternate_index: int = -1
    ) -> int:

        initial_index = self.get_container_stack_size() - 1
        container_index = alternate_index
        last_closed_container_info = self.get_closed_container_info(-1)

        # assert not last_closed_container_info.adjustment
        if last_closed_container_info.adjustment:
            adjust = 2 if container_index >= 0 else 1
        else:
            adjust = self.__calculate_adjust(initial_index, container_index)
        # endif

        index = (
            LeadingSpaceIndexTracker.calculate_token_line_number(token)
            - self.get_container_stack_item(initial_index).line_number
        )
        index -= (
            last_closed_container_info.adjustment
            + last_closed_container_info.list_nudge_count
        )
        index -= adjust
        return index

    def __calculate_adjust(self, initial_index: int, container_index: int) -> int:

        last_closed_container_info = self.get_closed_container_info(-1)
        assert (
            initial_index < 2
            or container_index
            or not last_closed_container_info.adjustment
        )
        return (
            0
            if initial_index >= 1
            and not container_index
            and last_closed_container_info.adjustment
            else 1
        )



"""
Module to provide details about a plugin, supplied by the plugin.
"""

from dataclasses import dataclass
from typing import Optional, Union


# pylint: disable=too-many-instance-attributes
@dataclass(frozen=True)
class PluginDetails:
    """
    Class to provide details about a plugin, supplied by the plugin.
    """

    plugin_id: str
    plugin_name: str
    plugin_description: str
    plugin_enabled_by_default: bool
    plugin_version: str
    plugin_interface_version: int = 1
    plugin_url: Optional[str] = None
    plugin_configuration: Optional[str] = None


# pylint: enable=too-many-instance-attributes


@dataclass(frozen=True)
class PluginDetailsV2(PluginDetails):
    """
    Class to provide details about a plugin, supplied by the plugin.
    """

    plugin_supports_fix: bool = False
    plugin_interface_version: int = 2
    plugin_fix_level: int = 1


@dataclass(frozen=True)
class PluginDetailsV3(PluginDetailsV2):
    """
    Class to provide details about a plugin, supplied by the plugin.
    """

    plugin_interface_version: int = 3


@dataclass(frozen=True)
class QueryConfigItem:
    """
    Class to provide information on the configuration items for the plugin.
    """

    name: str
    value: Union[bool, int, str]



"""
Module to implement a plugin that looks for code blocks for scripts that do not
look right.
"""

from typing import cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd014(RulePlugin):
    """
    Class to implement a plugin that looks for code blocks for scripts that do not
    look right.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__in_code_block = False

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="commands-show-output",
            plugin_id="MD014",
            plugin_enabled_by_default=True,
            plugin_description="Dollar signs used before commands without showing output",
            plugin_version="0.5.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md014.md",
            plugin_supports_fix=False,
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__in_code_block = False

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_code_block:
            self.__in_code_block = True
        elif token.is_code_block_end:
            self.__in_code_block = False
        elif self.__in_code_block and token.is_text:
            text_token = cast(TextMarkdownToken, token)
            split_token_text = text_token.token_text.split(
                ParserHelper.newline_character
            )
            if all(
                next_line.strip(" ").startswith("$") for next_line in split_token_text
            ):
                self.report_next_token_error(context, token)



"""
Module to specifically deal with handling the file scanning
for the application.
"""

import argparse
import copy
import logging
import os
import shutil
import sys
import tempfile
from typing import Callable, Dict, List, Optional, Set, Tuple, Union, cast

from pymarkdown.application_file_scanner import ApplicationFileScanner
from pymarkdown.extensions.pragma_token import PragmaToken
from pymarkdown.general.bad_tokenization_error import BadTokenizationError
from pymarkdown.general.main_presentation import MainPresentation
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.source_providers import FileSourceProvider
from pymarkdown.general.tokenized_markdown import TokenizedMarkdown
from pymarkdown.plugin_manager.bad_plugin_error import BadPluginError
from pymarkdown.plugin_manager.bad_plugin_fix_error import BadPluginFixError
from pymarkdown.plugin_manager.fix_line_record import FixLineRecord
from pymarkdown.plugin_manager.fix_token_record import FixTokenRecord
from pymarkdown.plugin_manager.found_plugin import FoundPlugin
from pymarkdown.plugin_manager.plugin_manager import PluginManager
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.replace_tokens_record import ReplaceTokensRecord
from pymarkdown.return_code_helper import ApplicationResult, ReturnCodeHelper
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_markdown.transform_to_markdown import TransformToMarkdown

POGGER = ParserLogger(logging.getLogger(__name__))


class FileScanHelper:
    """
    Class to specifically deal with handling the file scanning
    for the application.
    """

    __normal_scan_subcommand = "scan"
    __stdin_scan_subcommand = "scan-stdin"
    __normal_fix_subcommand = "fix"

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        tokenizer: TokenizedMarkdown,
        plugins: PluginManager,
        presentation: MainPresentation,
        show_stack_trace: bool,
        handle_error: Callable[[str, Optional[Exception], bool, str], None],
    ):
        self.__tokenizer = tokenizer
        self.__plugins = plugins
        self.__presentation = presentation
        self.__show_stack_trace = show_stack_trace
        self.__handle_error = handle_error
        self.__continue_on_error = False

    # pylint: enable=too-many-arguments

    def process_files_to_scan(
        self,
        args: argparse.Namespace,
        use_standard_in: bool,
        files_to_scan: List[str],
        string_to_scan: Optional[str],
    ) -> Tuple[bool, bool]:
        """
        Process the specified files to scan, or the string to scan.
        """

        self.__continue_on_error = args.continue_on_error
        in_fix_mode = args.primary_subparser == FileScanHelper.__normal_fix_subcommand

        # sourcery skip: raise-specific-error
        did_fix_any_file = False
        did_fail_any_file = False
        if use_standard_in:
            assert not in_fix_mode, "Standard-in cannot be used with fix mode."
            POGGER.debug("Scanning from: (stdin)")
            self.__scan_from_stdin(args, string_to_scan)

        else:
            POGGER.debug("Scanning from: $", files_to_scan)
            for next_file in files_to_scan:
                if in_fix_mode:
                    did_fix_file, did_succeed = self.__fix_specific_file(
                        next_file,
                        next_file,
                        args.x_fix_debug,
                        args.x_fix_file_debug,
                        args.x_fix_no_rescan_log,
                    )
                    if did_fix_file:
                        self.__presentation.print_fix_message(next_file)
                        did_fix_any_file = True
                else:
                    did_succeed = self.__scan_specific_file(next_file, next_file)
                if not did_succeed:
                    did_fail_any_file = True
        return did_fix_any_file, did_fail_any_file

    def __scan_from_stdin(
        self, args: argparse.Namespace, string_to_scan: Optional[str]
    ) -> None:
        temporary_file = None
        scan_exception = None
        scan_id = "stdin" if string_to_scan is None else "in-memory"
        try:
            if args.x_test_stdin_fault:
                raise IOError("made up")

            with tempfile.NamedTemporaryFile("wt", delete=False) as outfile:
                temporary_file = outfile.name

                if string_to_scan:
                    outfile.write(string_to_scan)
                else:
                    for line in sys.stdin:
                        outfile.write(line)

            self.__scan_specific_file(temporary_file, scan_id)

        except IOError as this_exception:
            scan_exception = this_exception
        finally:
            if temporary_file and os.path.exists(temporary_file):
                os.remove(temporary_file)

        if scan_exception:
            try:
                raise IOError(
                    f"Temporary file to capture {scan_id} was not written ({scan_exception})."
                ) from scan_exception
            except IOError as this_exception:
                self.__handle_scan_error(scan_id, this_exception)

    def __scan_specific_file(self, next_file: str, next_file_name: str) -> bool:
        try:
            source_provider = FileSourceProvider(next_file)
            self.__scan_file(source_provider, next_file_name)
            return True
        except BadPluginError as this_exception:
            self.__handle_scan_error(next_file, this_exception, allow_shortcut=True)
        except BadTokenizationError as this_exception:
            if not self.__continue_on_error:
                raise
            self.__handle_scan_error(next_file, this_exception, allow_shortcut=True)
        return False

    def __scan_file(
        self, source_provider: FileSourceProvider, next_file_name: str
    ) -> None:  # sourcery skip: extract-method
        """
        Scan a given file and call the plugin manager for any significant events.
        """

        POGGER.info("Scanning file '$'.", next_file_name)
        context = None

        try:
            POGGER.info("Starting file '$'.", next_file_name)

            POGGER.info("Scanning file '$' token-by-token.", next_file_name)
            actual_tokens = self.__tokenizer.transform_from_provider(
                source_provider, do_add_end_of_stream_token=True
            )
            context = self.__plugins.starting_new_file(next_file_name, actual_tokens)

            source_provider.reset_to_start()
            self.__process_file_scan(
                context, source_provider, next_file_name, actual_tokens
            )

            context.report_on_triggered_rules()
            POGGER.info("Ending file '$'.", next_file_name)
        except Exception:
            if context:
                context.report_on_triggered_rules()
            POGGER.info("Ending file '$' with exception.", next_file_name)
            raise

    def __process_file_scan(
        self,
        context: PluginScanContext,
        source_provider: FileSourceProvider,
        next_file_name: str,
        actual_tokens: List[MarkdownToken],
    ) -> None:
        if actual_tokens and actual_tokens[-1].is_pragma:
            pragma_token = cast(PragmaToken, actual_tokens[-1])
            self.__plugins.compile_pragmas(next_file_name, pragma_token.pragma_lines)
            actual_tokens = actual_tokens[:-1]

        POGGER.info("Scanning file '$' tokens.", next_file_name)
        for next_token in actual_tokens:
            POGGER.info("Processing token: $", next_token)
            self.__plugins.next_token(context, next_token)

        POGGER.info("Completed scanning tokens in file '$'.", next_file_name)

        POGGER.info("Scanning file '$' line-by-line.", next_file_name)
        self.__process_lines_in_file(source_provider, context, next_file_name)

    # pylint: disable=too-many-arguments
    def __fix_specific_file(
        self,
        next_file: str,
        next_file_name: str,
        fix_debug: bool,
        fix_file_debug: bool,
        fix_nolog_rescan: bool,
    ) -> Tuple[bool, bool]:
        did_fix_file = False
        did_succeed = False
        try:
            try:
                POGGER.info("Starting file to fix '$'.", next_file_name)

                did_fix_file = self.__process_file_fix(
                    next_file,
                    next_file_name,
                    fix_debug,
                    fix_file_debug,
                    fix_nolog_rescan,
                )

                POGGER.info("Ending file to fix '$'.", next_file_name)
                did_succeed = True
            except Exception:
                POGGER.info("Ending file to fix '$' with exception.", next_file_name)
                raise
        except (BadPluginError, BadPluginFixError) as this_exception:
            self.__handle_scan_error(next_file, this_exception, allow_shortcut=True)
        except BadTokenizationError as this_exception:
            if not self.__continue_on_error:
                raise
            self.__handle_scan_error(next_file, this_exception, allow_shortcut=True)
        return did_fix_file, did_succeed

    # pylint: enable=too-many-arguments

    def __handle_scan_error(
        self, next_file: str, this_exception: Exception, allow_shortcut: bool = False
    ) -> None:
        if not self.__continue_on_error:
            allow_shortcut = False

        if allow_shortcut:
            show_extended_information = False
            print_prefix = ""
        else:
            show_extended_information = self.__show_stack_trace
            print_prefix = "\n\n"

        if formatted_error := self.__presentation.format_scan_error(
            next_file, this_exception, show_extended_information, allow_shortcut
        ):
            self.__handle_error(
                formatted_error, this_exception, not allow_shortcut, print_prefix
            )

        # If the `format_scan_error` call above returned None, it meant that
        # it handled any required output.  However, the application still needs
        # to terminate to respect that the error was called.
        #
        # update: the allow_shortcut variable allows for a shorter error to be
        #   logged to allow processing to continue.  When that happens, the
        #   application is not exitted.
        if not allow_shortcut:
            ReturnCodeHelper.exit_application(ApplicationResult.SYSTEM_ERROR)

    def __process_file_fix_rescan(
        self, fix_debug: bool, fix_nolog_rescan: bool, next_file_two: str
    ) -> List[MarkdownToken]:
        _ = fix_debug
        # if fix_debug:
        #     print(f"scan: {next_file_two}")
        POGGER.info("Rescanning file '$' before line-by-line fixes.", next_file_two)
        source_provider = FileSourceProvider(next_file_two)

        if fix_nolog_rescan:
            saved_log_level = logging.WARNING
            if POGGER.is_enabled_for(logging.DEBUG):
                saved_log_level = logging.DEBUG
            elif POGGER.is_enabled_for(logging.INFO):
                saved_log_level = logging.INFO
            logging.getLogger().setLevel(logging.WARNING)
            ParserLogger.sync_on_next_call()
            logging.getLogger().warning(
                "Setting logging level to WARN during rescan upon request."
            )

        try:
            actual_tokens = self.__tokenizer.transform_from_provider(
                source_provider, do_add_end_of_stream_token=True
            )
        finally:
            if fix_nolog_rescan:
                saved_log_level_name = logging.getLevelName(saved_log_level)
                logging.getLogger().warning(
                    "Restoring log level to %s.", saved_log_level_name
                )
                logging.getLogger().setLevel(saved_log_level)
                ParserLogger.sync_on_next_call()
        return actual_tokens

    # pylint: disable=too-many-arguments, too-many-locals
    def __process_file_fix_pass(
        self,
        next_file: str,
        next_file_name: str,
        fix_debug: bool,
        fix_file_debug: bool,
        fix_nolog_rescan: bool,
        fix_list: List[str],
        collect_list: List[str],
    ) -> Tuple[bool, Set[str], Set[str]]:
        # Scan the provided file for any token fixes.
        (
            next_file_two,
            actual_tokens,
            did_any_tokens_get_fixed,
            collected_token_triggers,
        ) = self.__process_file_fix_tokens(
            next_file, next_file_name, fix_debug, fix_file_debug, fix_list, collect_list
        )

        # If tokens are returned, then no changes were made due to tokens and the
        # tokenized list can be reused without any worry of changes.
        if actual_tokens:
            assert (
                not did_any_tokens_get_fixed
            ), "Reusing tokens assumes that no tokens were fixed/changed."
            POGGER.info(
                "Scanning for token fixes did not change file.  Reusing tokens for line-by-line fixes."
            )
        else:
            actual_tokens = self.__process_file_fix_rescan(
                fix_debug, fix_nolog_rescan, next_file_two
            )

        # As the lines are processed, a new temporary line file is written to. If either
        # tokens were fixed or lines were fixed, the file contents of the file
        # temporary_line_file_name will contain the updated document.
        (
            this_file_fix_line_records,
            temporary_line_file_name,
            collected_line_triggers,
        ) = self.__process_file_fix_lines(
            next_file_two,
            next_file_name,
            actual_tokens,
            fix_debug,
            fix_file_debug,
            fix_list,
            collect_list,
        )

        # If anything was fixed, copy the temporary file on top of the original file
        # that was scanned.
        did_any_lines_get_fixed = bool(this_file_fix_line_records)
        did_anything_get_fixed = did_any_lines_get_fixed or did_any_tokens_get_fixed
        if did_anything_get_fixed:
            if fix_debug and fix_file_debug:
                print(f"Copy {temporary_line_file_name} to {next_file}")
            shutil.copyfile(temporary_line_file_name, next_file)
        if fix_debug and fix_file_debug:
            print(f"Remove:{temporary_line_file_name}")
        os.remove(temporary_line_file_name)
        if next_file_two != next_file:
            if fix_debug and fix_file_debug:
                print(f"Remove:{next_file_two}")
            os.remove(next_file_two)

        return did_anything_get_fixed, collected_token_triggers, collected_line_triggers

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments, too-many-locals
    def __process_file_fix_next_level(
        self,
        plugins_by_fix_level: Dict[int, List[str]],
        minimum_fix_level: int,
        fixes_by_id: Dict[str, FoundPlugin],
        next_file: str,
        next_file_name: str,
        fix_debug: bool,
        fix_file_debug: bool,
        fix_nolog_rescan: bool,
    ) -> Tuple[bool, bool, int]:
        keep_processing = False
        collect_list = []
        fix_list = []
        for fix_level, level_list in plugins_by_fix_level.items():
            if fix_level == minimum_fix_level:
                fix_list = level_list[:]
            elif fix_level > minimum_fix_level:
                collect_list.extend(level_list)
        assert fix_list is not None

        (
            did_anything_get_fixed_this_time,
            collected_token_triggers,
            collected_line_triggers,
        ) = self.__process_file_fix_pass(
            next_file,
            next_file_name,
            fix_debug,
            fix_file_debug,
            fix_nolog_rescan,
            fix_list,
            collect_list,
        )

        trigger_set = collected_token_triggers | collected_line_triggers
        new_minimum_fix_level: Optional[int] = None
        for next_triggered_plugin_id in trigger_set:
            triggered_plugin_fix_level = fixes_by_id[
                next_triggered_plugin_id
            ].plugin_fix_level
            assert (
                triggered_plugin_fix_level > minimum_fix_level
            ), "Triggered plugin levels should only increase."
            new_minimum_fix_level = (
                triggered_plugin_fix_level
                if new_minimum_fix_level is None
                else min(new_minimum_fix_level, triggered_plugin_fix_level)
            )
        if new_minimum_fix_level is not None:
            keep_processing = True
            minimum_fix_level = new_minimum_fix_level

        return keep_processing, did_anything_get_fixed_this_time, minimum_fix_level

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments, too-many-locals
    def __process_file_fix(
        self,
        next_file: str,
        next_file_name: str,
        fix_debug: bool,
        fix_file_debug: bool,
        fix_nolog_rescan: bool,
    ) -> bool:
        enabled_plugins_with_fixes = filter(
            lambda x: x.plugin_supports_fix, self.__plugins.enabled_plugins
        )
        fixes_by_id: Dict[str, FoundPlugin] = {
            i.plugin_id.lower(): i for i in enabled_plugins_with_fixes
        }

        enabled_plugins_with_fixes = filter(
            lambda x: x.plugin_supports_fix, self.__plugins.enabled_plugins
        )
        fix_plugins_with_levels = [
            (i.plugin_id, i.plugin_fix_level) for i in enabled_plugins_with_fixes
        ]

        plugins_by_fix_level: Dict[int, List[str]] = {}
        for next_pair in fix_plugins_with_levels:
            pair_plugin_id = next_pair[0]
            pair_fix_level = next_pair[1]
            if pair_fix_level in plugins_by_fix_level:
                level_list = plugins_by_fix_level[pair_fix_level]
            else:
                level_list = []
                plugins_by_fix_level[pair_fix_level] = level_list
            level_list.append(pair_plugin_id)
        minimum_fix_level = min(plugins_by_fix_level.keys())
        did_anything_get_fixed = False
        keep_processing = True

        while keep_processing:
            (
                keep_processing,
                did_anything_get_fixed_this_time,
                minimum_fix_level,
            ) = self.__process_file_fix_next_level(
                plugins_by_fix_level,
                minimum_fix_level,
                fixes_by_id,
                next_file,
                next_file_name,
                fix_debug,
                fix_file_debug,
                fix_nolog_rescan,
            )
            did_anything_get_fixed = (
                did_anything_get_fixed or did_anything_get_fixed_this_time
            )

        return did_anything_get_fixed

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments, too-many-locals
    def __process_file_fix_lines(
        self,
        next_file: str,
        next_file_name: str,
        actual_tokens: List[MarkdownToken],
        fix_debug: bool,
        fix_file_debug: bool,
        fix_list: List[str],
        collect_list: List[str],
    ) -> Tuple[List[FixLineRecord], str, Set[str]]:
        source_provider = FileSourceProvider(next_file)
        with tempfile.NamedTemporaryFile() as temp_output:
            temporary_file_name = temp_output.name
        with open(temporary_file_name, "wt", encoding="utf-8") as source_file:
            POGGER.info("Scanning before line-by-line fixes.")
            fix_context = self.__plugins.starting_new_file(
                next_file_name,
                actual_tokens,
                fix_mode=True,
                temp_output=source_file,
                fix_token_map=None,
            )
            report_context = self.__plugins.starting_new_file(
                next_file_name, actual_tokens, constraint_id_list=collect_list
            )
            context_map: Dict[str, PluginScanContext] = {
                i: fix_context for i in fix_list
            }
            for i in collect_list:
                context_map[i] = report_context

            # Due to context required to process the line requirements, we need go
            # through all the tokens first, before processing the lines.
            #
            # Basically, to allow any of the rules to build context applicable to
            # the line being scanned, we rescan the tokens to present an updated
            # picture of the tokens.
            for next_token in actual_tokens:
                POGGER.info("Processing tokens: $", next_token)
                self.__plugins.next_token(fix_context, next_token, context_map)

            POGGER.info("Completed token scanning.")
            self.__process_lines_in_file(
                source_provider, fix_context, next_file_name, context_map
            )
            this_file_fix_line_records = fix_context.fix_line_records
            if this_file_fix_line_records and fix_debug:
                for next_record in fix_context.fix_line_records:
                    print(next_record)

        self.__print_file_in_debug_mode(fix_debug, fix_file_debug, temporary_file_name)
        return (
            this_file_fix_line_records,
            temporary_file_name,
            report_context.get_triggered_rules(),
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments, too-many-locals
    def __process_file_fix_tokens(
        self,
        next_file: str,
        next_file_name: str,
        fix_debug: bool,
        fix_file_debug: bool,
        fix_list: List[str],
        collect_list: List[str],
    ) -> Tuple[str, List[MarkdownToken], bool, Set[str]]:
        self.__print_file_in_debug_mode(fix_debug, fix_file_debug, next_file)

        POGGER.info("Scanning file to fix '$' token-by-token.", next_file_name)
        source_provider = FileSourceProvider(next_file)
        actual_tokens = self.__tokenizer.transform_from_provider(
            source_provider, do_add_end_of_stream_token=True
        )

        fix_token_map: Dict[MarkdownToken, List[FixTokenRecord]] = {}
        replace_tokens_list: List[ReplaceTokensRecord] = []
        fix_context = self.__plugins.starting_new_file(
            next_file_name,
            actual_tokens,
            fix_mode=True,
            temp_output=None,
            fix_token_map=fix_token_map,
            constraint_id_list=fix_list,
            replace_tokens_list=replace_tokens_list,
        )
        report_context = self.__plugins.starting_new_file(
            next_file_name, actual_tokens, constraint_id_list=collect_list
        )

        context_map = {i: fix_context for i in fix_list}
        for i in collect_list:
            context_map[i] = report_context

        for next_token in actual_tokens:
            POGGER.info("Processing token: $", next_token)
            self.__plugins.next_token(fix_context, next_token, context_map)

        POGGER.info("Completed scanning file '$' for fixes.", next_file_name)
        self.__plugins.completed_file(fix_context, -1, context_map)

        did_any_tokens_get_fixed = False
        if fix_context.get_fix_token_map() or fix_context.get_replace_tokens_list():
            (
                next_file,
                actual_tokens,
                did_any_tokens_get_fixed,
            ) = self.__process_file_fix_tokens_apply_fixes(
                fix_context,
                next_file,
                actual_tokens,
                fix_token_map,
                fix_debug,
                fix_file_debug,
                replace_tokens_list,
            )
        return (
            next_file,
            actual_tokens,
            did_any_tokens_get_fixed,
            report_context.get_triggered_rules(),
        )

    # pylint: enable=too-many-arguments, too-many-locals

    def __look_for_collisions(
        self,
        next_replacement: ReplaceTokensRecord,
        actual_tokens: List[MarkdownToken],
        fixed_token_indices: Dict[int, List[str]],
        replaced_token_indices: Dict[int, str],
    ) -> None:
        start_index = actual_tokens.index(next_replacement.start_token)
        end_index = actual_tokens.index(next_replacement.end_token)
        next_match_with_fixed_tokens = next(
            (i for i in range(start_index, end_index + 1) if i in fixed_token_indices),
            None,
        )
        if next_match_with_fixed_tokens is not None:
            previous_plugin_ids = ",".join(
                fixed_token_indices[next_match_with_fixed_tokens]
            )
            POGGER.info(
                "Previous plugins ($) set one or more fields for the current token.",
                previous_plugin_ids,
            )
            POGGER.info(
                "Current plugin ($) wants to replace the token with another token.",
                next_replacement.plugin_id,
            )

            raise BadPluginFixError(
                f"Multiple plugins ({previous_plugin_ids} and {next_replacement.plugin_id}) are in conflict about fixing the token."
            )
        for next_index in range(start_index, end_index + 1):
            if next_index in replaced_token_indices:
                POGGER.info(
                    "Previous plugin ($) replaced the token.",
                    replaced_token_indices[next_index],
                )
                POGGER.info(
                    "Current plugin ($) wants to replace the same token with another token.",
                    next_replacement.plugin_id,
                )

                raise BadPluginFixError(
                    f"Multiple plugins ({replaced_token_indices[next_index]} and {next_replacement.plugin_id}) are in conflict about replacing the token."
                )
            replaced_token_indices[next_index] = next_replacement.plugin_id

    def __apply_replacement_fix(
        self,
        context: PluginScanContext,
        next_replacement: ReplaceTokensRecord,
        actual_tokens: List[MarkdownToken],
    ) -> None:
        start_index = actual_tokens.index(next_replacement.start_token)
        end_index = actual_tokens.index(next_replacement.end_token)
        actual_start_index = start_index
        while (
            actual_start_index < end_index
            and actual_tokens[actual_start_index].is_end_token
        ):
            actual_start_index += 1
        line_number_delta_1 = (
            next_replacement.end_token.line_number
            - actual_tokens[actual_start_index].line_number
        ) + 1
        line_number_delta_2 = (
            next_replacement.replacement_tokens[-1].line_number
            - next_replacement.replacement_tokens[0].line_number
        ) + 1
        line_number_delta = line_number_delta_2 - line_number_delta_1

        new_tokens = actual_tokens[:start_index]
        new_tokens.extend(next_replacement.replacement_tokens)
        end_tokens = actual_tokens[end_index + 1 :]
        for next_token in end_tokens:
            next_token.adjust_line_number(context, line_number_delta)
        new_tokens.extend(end_tokens)

        if new_tokens[-1].is_pragma:
            pragma_token = cast(PragmaToken, new_tokens[-1])
            for pragma_line_number in sorted(pragma_token.pragma_lines.keys())[::-1]:
                if pragma_line_number > next_replacement.end_token.line_number:
                    pragma_token.adjust_pragma_line_number(
                        pragma_line_number, pragma_line_number + line_number_delta
                    )

        actual_tokens.clear()
        actual_tokens.extend(new_tokens)

    # pylint: disable=too-many-arguments
    def __apply_replacements(
        self,
        fix_debug: bool,
        context: PluginScanContext,
        did_any_tokens_get_fixed: bool,
        replace_tokens_list: List[ReplaceTokensRecord],
        actual_tokens: List[MarkdownToken],
        fixed_token_indices: Dict[int, List[str]],
        replaced_token_indices: Dict[int, str],
    ) -> bool:
        for next_replace_index in replace_tokens_list:
            self.__look_for_collisions(
                next_replace_index,
                actual_tokens,
                fixed_token_indices,
                replaced_token_indices,
            )

        # did_fix = False
        if fix_debug:
            print("TOKENS-----")
            for i, j in enumerate(actual_tokens):
                print(f" {i:02}:{ParserHelper.make_value_visible(j)}")
            print("TOKENS-----")
        for next_replace_index in replace_tokens_list:
            did_any_tokens_get_fixed = True
            # if fix_debug and not did_fix:
            #     did_fix = True
            #     print("BEFORE-XXX-----")
            #     for i,j in enumerate(actual_tokens):
            #         print(f" {i:02}:{ParserHelper.make_value_visible(j)}")
            #     print("BEFORE-XXX-----")
            if fix_debug:
                print(f" {ParserHelper.make_value_visible(next_replace_index)}")
            self.__apply_replacement_fix(context, next_replace_index, actual_tokens)
        if fix_debug:
            print("TOKENS-----")
            for i, j in enumerate(actual_tokens):
                print(f" {i:02}:{ParserHelper.make_value_visible(j)}")
            print("TOKENS-----")
        return did_any_tokens_get_fixed

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def __process_file_fix_tokens_apply_fixes(
        self,
        context: PluginScanContext,
        next_file: str,
        actual_tokens: List[MarkdownToken],
        fix_token_map: Dict[MarkdownToken, List[FixTokenRecord]],
        fix_debug: bool,
        fix_file_debug: bool,
        replace_tokens_list: List[ReplaceTokensRecord],
    ) -> Tuple[str, List[MarkdownToken], bool]:

        did_any_tokens_get_fixed = self.__process_file_fix_tokens_apply_fixes_inner(
            context, fix_debug, actual_tokens, replace_tokens_list, fix_token_map
        )

        transformer = TransformToMarkdown()
        markdown_from_tokens = transformer.transform(actual_tokens)

        if fix_debug:
            print(f"MARKDOWN:{ParserHelper.make_value_visible(markdown_from_tokens)}")
        with tempfile.NamedTemporaryFile() as temp_output:
            temporary_file_name = temp_output.name
        with open(temporary_file_name, "wt", encoding="utf-8") as source_file:
            source_file.write(markdown_from_tokens)
            next_file = temporary_file_name
            actual_tokens.clear()

        self.__print_file_in_debug_mode(fix_debug, fix_file_debug, temporary_file_name)
        return next_file, actual_tokens, did_any_tokens_get_fixed

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def __process_file_fix_tokens_apply_fixes_inner(
        self,
        context: PluginScanContext,
        fix_debug: bool,
        actual_tokens: List[MarkdownToken],
        replace_tokens_list: List[ReplaceTokensRecord],
        fix_token_map: Dict[MarkdownToken, List[FixTokenRecord]],
    ) -> bool:
        fixed_token_indices: Dict[int, List[str]] = {}
        replaced_token_indices: Dict[int, str] = {}

        did_any_tokens_get_fixed = False
        if fix_debug:
            print("--")
        for token_instance, requested_fixes in context.get_fix_token_map().items():
            if fix_debug:
                print(f"APPLY:{ParserHelper.make_value_visible(requested_fixes)}")
                print(f"BEFORE:{ParserHelper.make_value_visible(token_instance)}")
            self.__apply_token_fix(
                context, token_instance, requested_fixes, actual_tokens
            )
            actual_token_index = actual_tokens.index(token_instance)
            fixed_token_indices[actual_token_index] = [
                i.plugin_id for i in requested_fixes
            ]
            if fix_debug:
                print(f" AFTER:{ParserHelper.make_value_visible(token_instance)}")
        if fix_debug:
            print("--")
        did_any_tokens_get_fixed = self.__apply_replacements(
            fix_debug,
            context,
            did_any_tokens_get_fixed,
            replace_tokens_list,
            actual_tokens,
            fixed_token_indices,
            replaced_token_indices,
        )
        if fix_debug:
            print("--")

        did_any_tokens_get_fixed = did_any_tokens_get_fixed or bool(fix_token_map)
        return did_any_tokens_get_fixed

    # pylint: enable=too-many-arguments

    def __print_file_in_debug_mode(
        self, fix_debug: bool, fix_file_debug: bool, next_file: str
    ) -> None:
        if fix_debug and fix_file_debug:
            with open(next_file, "rt", encoding="utf-8") as source_file:
                file_contents = source_file.read()
                print(
                    "\n--"
                    + next_file
                    + "--\n"
                    + file_contents.replace("\n", "\\n")
                    + "\n--"
                )

    def __apply_token_fix(
        self,
        context: PluginScanContext,
        token_instance: MarkdownToken,
        requested_fixes: List[FixTokenRecord],
        actual_tokens: List[MarkdownToken],
    ) -> None:
        # TODO figure out if we have to copy and replace, or if modifying the token in place is enough.
        _ = actual_tokens

        # TODO Simple scan for now, in future, want to check and allow 2 names that set to the same value.
        fix_map: Dict[str, Tuple[Union[str, int], str]] = {}
        for next_fix_record in requested_fixes:
            map_key = str(next_fix_record.token_to_fix) + str(
                next_fix_record.field_name
            )
            if map_key in fix_map:
                recorded_value = next_fix_record.field_value
                fix_value = fix_map[map_key]

                current_plugin_id = next_fix_record.plugin_id.upper()
                previous_plugin_id = fix_value[1].upper()

                POGGER.info(
                    "Previous plugin ($) wants to set field $ to '$'",
                    previous_plugin_id,
                    next_fix_record.field_name,
                    recorded_value,
                )
                POGGER.info(
                    "Current plugin ($) wants to set field $ to '$'",
                    current_plugin_id,
                    next_fix_record.field_name,
                    fix_value[0],
                )

                raise BadPluginFixError(
                    f"Multiple plugins ({current_plugin_id} and {previous_plugin_id}) have requested a fix for the same field of the same token."
                )
            fix_map[map_key] = (next_fix_record.field_value, next_fix_record.plugin_id)

        for next_fix_record in requested_fixes:
            before_instance = copy.deepcopy(token_instance)
            if not token_instance.modify_token(
                context, next_fix_record.field_name, next_fix_record.field_value
            ):
                POGGER.debug(
                    "Token $ fix for plugin $ in action $ on field $ failed.",
                    next_fix_record.plugin_id,
                    next_fix_record.plugin_action,
                    next_fix_record.field_name,
                )
                raise BadPluginFixError(
                    formatted_message=f"Plugin id '{next_fix_record.plugin_id.upper()}'s '{next_fix_record.plugin_action}' action requested a token adjustment to field '{next_fix_record.field_name}' that failed."
                )
            POGGER.debug(
                "Token fix for plugin $ in action $ on field $ succeeded.",
                next_fix_record.plugin_id,
                next_fix_record.plugin_action,
                next_fix_record.field_name,
            )
            POGGER.debug(
                "Token before <$> and after <$>.", before_instance, token_instance
            )

    def __process_lines_in_file(
        self,
        source_provider: FileSourceProvider,
        context: PluginScanContext,
        next_file_name: str,
        context_map: Optional[Dict[str, PluginScanContext]] = None,
    ) -> None:
        line_number, next_line = 1, source_provider.get_next_line()
        while next_line is not None:
            POGGER.info("Processing line $: $", line_number, next_line)
            self.__plugins.next_line(
                context,
                line_number,
                next_line,
                source_provider.is_at_end_of_file,
                source_provider.did_final_line_end_with_newline,
                context_map,
            )
            line_number += 1
            next_line = source_provider.get_next_line()

        POGGER.info("Completed scanning lines in file '$'.", next_file_name)
        self.__plugins.completed_file(context, line_number, context_map)

    @staticmethod
    def is_scan_stdin_specified(args: argparse.Namespace) -> bool:
        """
        Specifies whether scanning from stdin was specified.
        """
        return FileScanHelper.__stdin_scan_subcommand == args.primary_subparser  # type: ignore

    @staticmethod
    def add_argparse_subparser(subparsers: argparse._SubParsersAction, is_fix_mode: bool) -> None:  # type: ignore
        """
        Add the subparser for scanning.
        """
        subparser_action = "fix" if is_fix_mode else "scan"
        subparser_command = (
            FileScanHelper.__normal_fix_subcommand
            if is_fix_mode
            else FileScanHelper.__normal_scan_subcommand
        )

        new_sub_parser = subparsers.add_parser(
            subparser_command,
            help=f"{subparser_action} the Markdown files in any specified paths",
        )
        ApplicationFileScanner.add_default_command_line_arguments(
            new_sub_parser, ".md", "Markdown"
        )

        if not is_fix_mode:
            subparsers.add_parser(
                FileScanHelper.__stdin_scan_subcommand,
                help="scan the standard input as a Markdown file",
            )



"""
Module to provide for an encapsulation of the high level state of the parser.
"""

from __future__ import annotations

import copy
from typing import List, Optional, Tuple, cast

from typing_extensions import Protocol

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import (
    LinkDefinitionStackToken,
    ListStackToken,
    StackToken,
    TableBlockStackToken,
)


# pylint: disable=too-few-public-methods
class CloseOpenBlocksProtocol(Protocol):
    """
    Protocol to provide typing for the close open blocks function.
    """

    # pylint: disable=too-many-arguments
    def __call__(  # noqa: E704
        self,
        parser_state: ParserState,
        destination_array: Optional[List[MarkdownToken]] = None,
        only_these_blocks: Optional[List[type]] = None,
        include_block_quotes: bool = False,
        include_lists: bool = False,
        until_this_index: int = -1,
        caller_can_handle_requeue: bool = False,
        requeue_reset: bool = False,
        was_forced: bool = False,
    ) -> Tuple[List[MarkdownToken], Optional[RequeueLineInfo]]: ...  # pragma: no cover

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class HandleBlankLineProtocol(Protocol):
    """
    Protocol to provide typing for the blank line function.
    """

    def __call__(  # noqa: E704
        self,
        parser_state: ParserState,
        input_line: str,
        from_main_transform: bool,
        position_marker: Optional[PositionMarker] = None,
    ) -> Tuple[
        Optional[List[MarkdownToken]], Optional[RequeueLineInfo]
    ]: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


# pylint: disable=too-many-instance-attributes, too-many-public-methods
class ParserState:
    """
    Class to provide for an encapsulation of the high level state of the parser.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        token_stack: List[StackToken],
        token_document: List[MarkdownToken],
        close_open_blocks_fn: CloseOpenBlocksProtocol,
        handle_blank_line_fn: HandleBlankLineProtocol,
        parse_properties: ParseBlockPassProperties,
    ) -> None:
        (
            self.__token_stack,
            self.__token_document,
            self.__close_open_blocks_fn,
            self.__handle_blank_line_fn,
        ) = (token_stack, token_document, close_open_blocks_fn, handle_blank_line_fn)

        self.__no_para_start_if_empty: bool = False

        self.__last_block_quote_stack_token: Optional[StackToken] = None
        self.__last_block_quote_markdown_token_index: Optional[int] = None
        self.__copy_of_last_block_quote_markdown_token: Optional[
            BlockQuoteMarkdownToken
        ] = None
        self.__x1_token: Optional[ListStartMarkdownToken] = None
        self.__copy_of_x1_token: Optional[ListStartMarkdownToken] = None
        self.__x1_token_index: Optional[int] = -1

        self.__original_stack_depth: int = 0
        self.__original_document_depth: int = 0
        self.__original_line_to_parse: Optional[str] = None
        self.__same_line_container_tokens: Optional[List[MarkdownToken]] = None
        self.nested_list_start: Optional[ListStackToken] = None
        self.copy_of_token_stack: List[StackToken] = []
        self.block_copy: List[Optional[MarkdownToken]] = []
        self.parse_properties = parse_properties

    # pylint: enable=too-many-arguments

    @property
    def token_stack(self) -> List[StackToken]:
        """
        Stack array containing the currently active stack tokens.
        """
        return self.__token_stack

    @property
    def token_document(self) -> List[MarkdownToken]:
        """
        Document array containing any Markdown tokens.
        """
        return self.__token_document

    @property
    def close_open_blocks_fn(self) -> CloseOpenBlocksProtocol:
        """
        Function to handle the closing of blocks.
        """
        return self.__close_open_blocks_fn

    @property
    def handle_blank_line_fn(self) -> HandleBlankLineProtocol:
        """
        Function to handle blank lines.
        """
        return self.__handle_blank_line_fn

    @property
    def same_line_container_tokens(self) -> Optional[List[MarkdownToken]]:
        """
        State of the container tokens at the start of the leaf processing.
        """
        return self.__same_line_container_tokens

    @property
    def last_block_quote_stack_token(self) -> Optional[StackToken]:
        """
        Last block quote token.
        """
        return self.__last_block_quote_stack_token

    @property
    def last_block_quote_markdown_token_index(self) -> Optional[int]:
        """
        Index of the last block quote token.
        """
        return self.__last_block_quote_markdown_token_index

    @property
    def copy_of_last_block_quote_markdown_token(
        self,
    ) -> Optional[BlockQuoteMarkdownToken]:
        """
        Copy of the last block quote markdown token, before any changes.
        """
        return self.__copy_of_last_block_quote_markdown_token

    @property
    def original_line_to_parse(self) -> Optional[str]:
        """
        Line to parse, before any processing occurs.
        """
        return self.__original_line_to_parse

    @property
    def original_stack_depth(self) -> int:
        """
        Length of the token_stack array at the start of the line.
        """
        return self.__original_stack_depth

    @property
    def original_document_depth(self) -> int:
        """
        Length of the token_document array at the start of the line.
        """
        return self.__original_document_depth

    @property
    def no_para_start_if_empty(self) -> bool:
        """
        Whether to start a paragraph if the owning list item was empty.
        """
        return self.__no_para_start_if_empty

    @property
    def x1_token(self) -> Optional[ListStartMarkdownToken]:
        """
        TBD
        """
        return self.__x1_token

    @property
    def copy_of_x1_token(self) -> Optional[ListStartMarkdownToken]:
        """
        TBD
        """
        return self.__copy_of_x1_token

    @property
    def x1_token_index(self) -> Optional[int]:
        """
        TBD
        """
        return self.__x1_token_index

    def __abc_part_1(
        self, number_of_lines_to_requeue: int, stack_token: TableBlockStackToken
    ) -> None:
        last_list_index = -1
        last_list_token = stack_token.x1_token
        if last_list_token is None:
            last_list_index = self.find_last_list_block_on_stack()
            if last_list_index > 0:
                last_list_token = self.token_stack[
                    last_list_index
                ].matching_markdown_token

        if last_list_token:
            list_token = cast(ListStartMarkdownToken, last_list_token)
            if last_list_index != -1:
                prev_leading_space_count = -1
            else:
                prev_leading_spaces = cast(
                    ListStartMarkdownToken, stack_token.copy_of_x1_token
                ).leading_spaces
                prev_leading_space_count = (
                    prev_leading_spaces.count("\n") if prev_leading_spaces else -1
                )

            leading_space_count = (
                list_token.leading_spaces.count("\n")
                if list_token.leading_spaces
                else 0
            )
            while (
                number_of_lines_to_requeue > 0
                and leading_space_count > prev_leading_space_count
            ):
                if list_token.leading_spaces:
                    list_token.remove_last_leading_space()
                number_of_lines_to_requeue -= 1
                leading_space_count = (
                    list_token.leading_spaces.count("\n")
                    if list_token.leading_spaces
                    else 0
                )

    def __abc_part_2(
        self,
        number_of_lines_to_requeue: int,
        requeue_line_info: RequeueLineInfo,
        stack_token: StackToken,
        last_block_stack_token: StackToken,
    ) -> None:
        if (
            number_of_lines_to_requeue > 1
            and not requeue_line_info.lines_to_requeue[0]
            and isinstance(stack_token, LinkDefinitionStackToken)
        ):
            number_of_lines_to_requeue -= 1

        assert last_block_stack_token.matching_markdown_token is not None
        # if last_block_stack_token.matching_markdown_token is not None:
        block_quote_token = cast(
            BlockQuoteMarkdownToken, last_block_stack_token.matching_markdown_token
        )
        # endif
        copy_of_block_quote_token = cast(
            TableBlockStackToken, stack_token
        ).copy_of_last_block_quote_markdown_token
        bleading_space_count = (
            copy_of_block_quote_token.bleading_spaces.count("\n")
            if copy_of_block_quote_token and copy_of_block_quote_token.bleading_spaces
            else -1
        )
        if isinstance(stack_token, TableBlockStackToken):
            bleading_space_count -= 1
        current_bleading_space_count = (
            block_quote_token.bleading_spaces.count("\n")
            if block_quote_token.bleading_spaces
            else 0
        )
        while (
            number_of_lines_to_requeue > 0
            and current_bleading_space_count > bleading_space_count
        ):
            if block_quote_token.bleading_spaces:
                block_quote_token.remove_last_bleading_space()
            number_of_lines_to_requeue -= 1
            current_bleading_space_count = (
                block_quote_token.bleading_spaces.count("\n")
                if block_quote_token.bleading_spaces
                else 0
            )

    def abc(self, requeue_line_info: RequeueLineInfo, stack_token: StackToken) -> None:
        """
        TBD

        need to have common "multi-line" stack token base, not assume TableBlockStackToken
        """
        if requeue_line_info.has_been_abc_ed:
            return
        modded_stack_token = cast(TableBlockStackToken, stack_token)
        number_of_lines_to_requeue = len(requeue_line_info.lines_to_requeue)
        self.__abc_part_1(number_of_lines_to_requeue, modded_stack_token)
        last_block_stack_token = cast(
            TableBlockStackToken, stack_token
        ).last_block_quote_stack_token
        if last_block_stack_token is None:
            if (last_block_quote_index := self.find_last_block_quote_on_stack()) > 0:
                last_block_stack_token = self.token_stack[last_block_quote_index]
        if last_block_stack_token:
            self.__abc_part_2(
                number_of_lines_to_requeue,
                requeue_line_info,
                stack_token,
                last_block_stack_token,
            )
        requeue_line_info.has_been_abc_ed = True

    def find_last_block_quote_on_stack(self) -> int:
        """
        Finds the index of the last block quote on the stack (from the end).
        If no block quotes are found, 0 is returned.
        """
        last_stack_index = len(self.token_stack) - 1
        while (
            not self.token_stack[last_stack_index].is_document
            and not self.token_stack[last_stack_index].is_block_quote
        ):
            last_stack_index -= 1
        return last_stack_index

    def find_last_list_block_on_stack(self) -> int:
        """
        Finds the index of the last list block on the stack (from the end).
        If no block quotes are found, 0 is returned.
        """
        last_stack_index = len(self.token_stack) - 1
        while (
            not self.token_stack[last_stack_index].is_document
            and not self.token_stack[last_stack_index].is_list
        ):
            last_stack_index -= 1
        return last_stack_index

    def find_last_container_on_stack(self) -> int:
        """
        Finds the index of the last container on the stack (from the end).
        If no container tokens are found, 0 is returned.
        """
        last_stack_index = len(self.token_stack) - 1
        while not self.token_stack[last_stack_index].is_document and not (
            self.token_stack[last_stack_index].is_block_quote
            or self.token_stack[last_stack_index].is_list
        ):
            last_stack_index -= 1
        return last_stack_index

    def count_of_block_quotes_on_stack(self) -> int:
        """
        Helper method to count the number of block quotes currently on the stack.
        """

        return sum(
            bool(next_item_on_stack.is_block_quote)
            for next_item_on_stack in self.token_stack
        )

    def mark_start_information(self, position_marker: PositionMarker) -> None:
        """
        Mark the start of processing this line of information.  A lot of
        this information is to allow a requeue to occur, if needed.
        """
        (
            self.__original_line_to_parse,
            self.__original_stack_depth,
            self.__original_document_depth,
            self.__no_para_start_if_empty,
        ) = (
            position_marker.text_to_parse[:],
            len(self.token_stack),
            len(self.token_document),
            False,
        )

        last_stack_index = self.find_last_block_quote_on_stack()

        (
            self.__last_block_quote_stack_token,
            self.__last_block_quote_markdown_token_index,
        ) = (None, None)
        self.__copy_of_last_block_quote_markdown_token = None
        if not self.token_stack[last_stack_index].is_document:
            self.__last_block_quote_stack_token = self.token_stack[last_stack_index]
            markdown_token = self.token_stack[last_stack_index].matching_markdown_token
            assert (
                markdown_token is not None
            ), "Always start with a container or leaf token, that has a matching markdown token."
            try:
                self.__last_block_quote_markdown_token_index = (
                    self.token_document.index(markdown_token)
                )
                self.__copy_of_last_block_quote_markdown_token = cast(
                    BlockQuoteMarkdownToken,
                    copy.deepcopy(
                        self.token_document[
                            self.__last_block_quote_markdown_token_index
                        ]
                    ),
                )
            except ValueError:
                self.__last_block_quote_markdown_token_index = -1
                self.__copy_of_last_block_quote_markdown_token = None
        x1 = self.find_last_list_block_on_stack()
        self.__x1_token = None
        self.__copy_of_x1_token = None
        self.__x1_token_index = -1
        if not self.token_stack[x1].is_document:
            matching_token = self.token_stack[x1].matching_markdown_token
            assert matching_token is not None
            assert isinstance(matching_token, ListStartMarkdownToken)
            self.__x1_token = matching_token

            self.__copy_of_x1_token = copy.deepcopy(self.__x1_token)
            self.__x1_token_index = self.token_document.index(self.__x1_token)

    def mark_for_leaf_processing(
        self, container_level_tokens: List[MarkdownToken]
    ) -> None:
        """
        Set things up for leaf processing.
        """
        self.__same_line_container_tokens = container_level_tokens

    def clear_after_leaf_processing(self) -> None:
        """
        Reset things after leaf processing.
        """
        self.__same_line_container_tokens = None

    def set_no_para_start_if_empty(self) -> None:
        """
        Set the member variable to true.
        """
        self.__no_para_start_if_empty = True


# pylint: enable=too-many-instance-attributes, too-many-public-methods
